# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logger = logging.getLogger(__name__)

import json

from lxml.html.builder import E

from neurons.form import THtmlFormRoot, SimpleRenderWidget
from neurons.polymer.model import PaperCheckbox, NeuronsArray, Template, \
    IronDataTableColumn, NeuronsDateTimePicker
from neurons.polymer.protocol.widget import PolymerWidgetBase

from spyne import ComplexModelBase, Unicode, Decimal, Boolean, DateTime, \
    Integer, AnyUri, Fault, D, Array, Date, Time
from spyne.util.cdict import cdict


_NA_COUNTER = 0


class PolymerForm(THtmlFormRoot(PolymerWidgetBase)):
    HTML_FORM = 'form'

    def __init__(self, app=None, ignore_uncap=False, ignore_wrappers=False,
                cloth=None, cloth_parser=None, polymorphic=True, hier_delim='.',
                     doctype=None, label=True, placeholder=None,
              action=None, method='POST', before_form=None, mrpc_url_prefix=''):

        super(PolymerForm, self).__init__(app=app, doctype=doctype,
                     ignore_uncap=ignore_uncap, ignore_wrappers=ignore_wrappers,
                cloth=cloth, cloth_parser=cloth_parser, polymorphic=polymorphic,
                    hier_delim=hier_delim, label=label, placeholder=placeholder,
                          action=action, method=method, before_form=before_form,
                                                mrpc_url_prefix=mrpc_url_prefix)

        self.serialization_handlers = cdict({
            Date: self._check_simple(self.date_to_parent),
            Time: self._check_simple(self.time_to_parent),
            # Uuid: self._check_simple(self.uuid_to_parent),
            # File: self.file_to_parent,
            Fault: self.fault_to_parent,
            AnyUri: self._check_simple(self.any_uri_to_parent),
            # AnyXml: self._check_simple(self.any_xml_to_parent),
            Integer: self._check_simple(self.integer_to_parent),
            Unicode: self._check_simple(self.unicode_to_parent),
            # AnyHtml: self._check_simple(self.any_html_to_parent),
            Decimal: self._check_simple(self.decimal_to_parent),
            Boolean: self._check_simple(self.boolean_to_parent),
            # Duration: self._check_simple(self.duration_to_parent),
            DateTime: self._check_simple(self.datetime_to_parent),
            ComplexModelBase: self.complex_model_to_parent,
        })

        self.hier_delim = hier_delim

        self.use_global_null_handler = False

        self.simple = SimpleRenderWidget(label=label)

    def _gen_form_attrib(self, ctx, cls):
        attrib = {
            'id': 'form',
            'is': 'iron-form',
            'method': self.method,
            'content-type': "application/json",
        }

        attrib.update(self._get_form_action(ctx, self.get_cls_attrs(cls)))

        return attrib

    def _write_submit_button(self, parent):
        parent.write(
            E.div(
                E.input(
                    value="Submit",
                    type="submit",
                    **{'class': 'submit'}
                ),
                E.div(**{"class$": "[[submitStatus]]"}),
                E.div("[[submitError]]", **{'class': 'submit-error'}),
            ),
        )

    # FIXME: this can only handle the simplest of cases.
    def array_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        global _NA_COUNTER

        cls_attrs = self.get_cls_attrs(cls)

        if issubclass(cls, Array):
            cls, = cls._type_info.values()

        fti = cls.get_flat_type_info(cls)

        columns = []

        for k, v in fti.items():
            subcls_attrs = self.get_cls_attrs(v)
            if subcls_attrs.exc:
                continue

            subname= self.trc(cls, ctx.locale, k)
            columns.append(IronDataTableColumn(
                name=subname, template=Template(data="[[item.%s]]" % (k,))
            ))

        methods = cls.Attributes.methods

        na_id = "na-%d" % _NA_COUNTER
        _NA_COUNTER += 1

        pam = {}

        if methods is not None and len(methods) > 0:
            mrpc_delim_elt = E.span(" | ")  # TODO: parameterize this
            first = True
            for mn, md in methods.items():
                if first:
                    first = False

                elif mrpc_delim_elt is not None:
                    parent.write(" ")
                    parent.write(mrpc_delim_elt)

                # get first argument name
                cls_arg_name = next(iter(md.in_message._type_info.keys()))

                pd = []
                for k, v in cls.get_identifiers():
                    pd.append("{0}.{1}=[[item.{1}]]".format(cls_arg_name, k))
                pd.append('[[_urlencodeParams("%s", "%s")]]' % (na_id, md.name))

                mdid2key = ctx.app.interface.method_descriptor_id_to_key
                method_key = mdid2key.get(id(md), None)
                if method_key is None:
                    continue

                href = method_key.rsplit("}", 1)[-1]
                href_with_params = "%s%s?%s" % \
                                      (self.mrpc_url_prefix, href, '&'.join(pd))

                text = md.translate(ctx.locale, md.in_message.get_type_name())

                if md.udd.parent_arg_map is not None:
                    pam[md.name] = md.udd.parent_arg_map

                columns.append(IronDataTableColumn(
                    template=Template(
                        data=E.a(text, **{'href$': href_with_params}),
                    )
                ))

        wgt_inst = NeuronsArray(
            id=na_id,
            name=self._gen_input_name(name),
            label=self.trc(cls, ctx.locale, name),
            columns=columns,
            arg_map=json.dumps(pam)
        )

        self._add_label(ctx, cls, cls_attrs, name, wgt_inst, **kwargs)
        self._write_elt_inst(ctx, wgt_inst, parent)

    def complex_model_to_parent(self, ctx, cls, inst, parent, name,
                                                      from_arr=False, **kwargs):
        cls_attr = self.get_cls_attrs(cls)

        if not from_arr and issubclass(cls, Array) or cls_attr.max_occurs > 1:
            return self.array_to_parent(ctx, cls, inst, parent, name, **kwargs)

        fieldset_attr = {'class': cls.get_type_name()}

        if cls_attr.fieldset is False or cls_attr.no_fieldset:
            return self._render_complex(ctx, cls, inst, parent, name, False,
                                                                       **kwargs)
        if cls_attr.fieldset or cls_attr.no_fieldset is False:
            with parent.element('fieldset', fieldset_attr):
                return self._render_complex(ctx, cls, inst, parent, name, False,
                                                                       **kwargs)

        with parent.element('fieldset', fieldset_attr):
            return self._render_complex(ctx, cls, inst, parent, name, True,
                                                                       **kwargs)

    def unicode_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        elt_inst = self._gen_widget_inst(ctx, cls, inst, name, cls_attrs,
                                                                       **kwargs)

        if cls_attrs.min_len is not None and cls_attrs.min_len > D('-inf'):
            elt_inst.minlength = cls_attrs.min_len

        if cls_attrs.max_len is not None and cls_attrs.max_len < D('inf'):
            elt_inst.maxlength = cls_attrs.max_len

        self._add_label(ctx, cls, cls_attrs, name, elt_inst, **kwargs)
        self._write_elt_inst(ctx, elt_inst, parent)

    def integer_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        elt_inst = self._gen_widget_inst(ctx, cls, inst, name, cls_attrs,
                                                                       **kwargs)
        elt_inst.type = "number"

        self._apply_number_constraints(cls_attrs, elt_inst, epsilon=1)

        self._add_label(ctx, cls, cls_attrs, name, elt_inst, **kwargs)
        self._write_elt_inst(ctx, elt_inst, parent)

    def decimal_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        elt_inst = self._gen_widget_inst(ctx, cls, inst, name, cls_attrs,
                                                                       **kwargs)
        elt_inst.type = "number"

        if D(cls.Attributes.fraction_digits).is_infinite():
            epsilon = self.HTML5_EPSILON
            elt_inst.step = 'any'

        else:
            epsilon = 10 ** (-int(cls.Attributes.fraction_digits))
            elt_inst.step = str(epsilon)

        self._apply_number_constraints(cls_attrs, elt_inst, epsilon=epsilon)

        self._add_label(ctx, cls, cls_attrs, name, elt_inst)
        self._write_elt_inst(ctx, elt_inst, parent)

    def date_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        elt_inst = self._gen_widget_inst(ctx, cls, inst, name, cls_attrs,
                                         sugcls=NeuronsDateTimePicker, **kwargs)
        elt_inst.no_time = True

        date_format = self._get_date_format(cls_attrs)
        if date_format is not None:
            # https://momentjs.com/docs/#/parsing/string-format/
            # TODO: handle all date tokens, raise on ones that can't be handled
            elt_inst.format = date_format \
                  .replace('%Y', 'YYYY').replace('%m', 'MM').replace('%d', 'DD')

        self._add_label(ctx, cls, cls_attrs, name, elt_inst)
        self._write_elt_inst(ctx, elt_inst, parent)

    def time_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        elt_inst = self._gen_widget_inst(ctx, cls, inst, name, cls_attrs,
                                         sugcls=NeuronsDateTimePicker, **kwargs)
        elt_inst.no_date = True

        time_format = self._get_time_format(cls_attrs)
        if time_format is not None:
            # https://momentjs.com/docs/#/parsing/string-format/
            # TODO: handle all time tokens, raise on ones that can't be handled
            elt_inst.format = time_format \
                    .replace('%H', 'HH').replace('%M', 'mm').replace('%S', 'ss')

        self._add_label(ctx, cls, cls_attrs, name, elt_inst)
        self._write_elt_inst(ctx, elt_inst, parent)

    def datetime_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        elt_inst = self._gen_widget_inst(ctx, cls, inst, name, cls_attrs,
                                         sugcls=NeuronsDateTimePicker, **kwargs)

        dt_format = self._get_datetime_format(cls_attrs)
        if dt_format is not None:
            # https://momentjs.com/docs/#/parsing/string-format/
            # TODO: handle all date/time tokens, raise on ones that can
            # not be handled
            elt_inst.format = dt_format \
                .replace('%Y', 'YYYY').replace('%m', 'MM').replace('%d', 'DD') \
                .replace('%H', 'HH').replace('%M', 'mm').replace('%S', 'ss')

        self._add_label(ctx, cls, cls_attrs, name, elt_inst)
        self._write_elt_inst(ctx, elt_inst, parent)

    def boolean_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        elt_inst = self._gen_widget_inst(ctx, cls, inst, name, cls_attrs,
                                                 sugcls=PaperCheckbox, **kwargs)

        self._write_elt_inst(ctx, elt_inst, parent)




import unittest

import yaml

from neurons.daemon import ServiceDaemon
from neurons import __version__ as NEURONS_VERSION

TEST_CONFIG = """
SomeClass:
    daemonize: false
    loggers:
    -   Logger:
            level: DEBUG
            path: .
    main_store: sql_main
    name: somedaemon
    secret: c29tZSBzZWNyZXQ=
    services:
    -   HttpServer:
            type: tcp4
            host: 0.0.0.0
            name: someservice
            port: 7001
            subapps:
            -   StaticFileServer:
                    disallowed_exts: []
                    list_contents: false
                    path: /home/plq/src/github/plq/neurons/assets
                    url: assets
    stores:
    -   Relational:
            async_pool: true
            backend: sqlalchemy
            conn_str: postgresql://postgres:@localhost:5432/somedaemon_plq
            echo_pool: false
            max_overflow: 3
            name: sql_main
            pool_recycle: 3600
            pool_size: 10
            pool_timeout: 30
            sync_pool: true
            sync_pool_type: QueuePool
    uuid: deac2096-0e18-11e6-a294-5453edabe249
"""


class TestConfig(unittest.TestCase):
    def test_migrate_versionless_to_versioned(self):
        ret = ServiceDaemon()._migrate_impl(TEST_CONFIG)
        retd = yaml.load(ret)
        key, = retd.keys()

        store = iter(retd.values()).next()['stores'][0]
        assert not ('Relational' in store)
        assert 'RelationalStore' in store
        assert len(retd) == 1
        assert retd[key]['file-version'] == NEURONS_VERSION

    def test_uidgid(self):
        config = ServiceDaemon.parse_config_string(TEST_CONFIG, "test")
        assert config.get_gid() == -1
        assert config.get_uid() == -1

        config.gid = "root"
        assert config.get_gid() == 0

        config.uid = "root"
        assert config.get_uid() == 0


if __name__ == '__main__':
    unittest.main()



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logger = logging.getLogger(__name__)

import re
import json

from spyne import ComplexModelBase, Boolean, Decimal
from spyne.util import six
from spyne.util.oset import oset
from spyne.util.cdict import cdict
from spyne.protocol.cloth import XmlCloth
from spyne.protocol.html import HtmlCloth

from neurons.form.widget import camel_case_to_uscore
from neurons.polymer.model import PaperInput, PaperDropdownMenu, PaperListbox, \
    PaperItem, NeuronsComplexDropdown, NeuronsComplexHref


class PolymerWidgetBase(HtmlCloth):
    # noinspection PyUnusedLocal
    def __init__(self, app=None, encoding='utf8',
                      mime_type=None, ignore_uncap=False, ignore_wrappers=False,
                                cloth=None, cloth_parser=None, polymorphic=True,
                              strip_comments=True, hier_delim='.', doctype=None,
                                                   placeholder=None, label=True,
               # the rest is not used, required for compat with HtmlFormWidget
               input_class=None, input_div_class=None, input_wrapper_class=None,
                                                              label_class=None):

        super(PolymerWidgetBase, self).__init__(app=app, encoding=encoding,
                                         doctype=doctype, hier_delim=hier_delim,
                                 mime_type=mime_type, ignore_uncap=ignore_uncap,
                                   ignore_wrappers=ignore_wrappers, cloth=cloth,
                             cloth_parser=cloth_parser, polymorphic=polymorphic,
                                                  strip_comments=strip_comments)

        self.label = label
        self.placeholder = placeholder
        self.use_global_null_handler = False

    @staticmethod
    def _apply_number_constraints(cls_attrs, elt_inst, epsilon):
        if cls_attrs.max_str_len != Decimal.Attributes.max_str_len:
            elt_inst.maxlength = cls_attrs.max_str_len

        if cls_attrs.ge != Decimal.Attributes.ge:
            elt_inst.min = cls_attrs.ge

        if cls_attrs.gt != Decimal.Attributes.gt:
            elt_inst.min = cls_attrs.gt + epsilon

        if cls_attrs.le != Decimal.Attributes.le:
            elt_inst.max = cls_attrs.le

        if cls_attrs.lt != Decimal.Attributes.lt:
            elt_inst.max = cls_attrs.lt - epsilon

        if elt_inst.min is None and cls_attrs.min_bound is not None:
            elt_inst.min = str(cls_attrs.min_bound)

        if elt_inst.max is None and cls_attrs.max_bound is not None:
            elt_inst.max = str(cls_attrs.max_bound)

    def _write_elt_inst(self, ctx, elt_inst, parent):
        elt_cls = elt_inst.__class__
        XmlCloth().to_parent(ctx, elt_cls, elt_inst, parent,
                                      elt_cls.Attributes.sub_name, use_ns=False)
        parent.write("\n")

    def _gen_elt_id(self, name, sugcls, array_index=None, **kwargs):
        wigname = camel_case_to_uscore(sugcls.get_type_name())
        if array_index is None:
            return "%s_%s" % (self.selsafe(name), wigname)
        return "%s_%d_%s" % (self.selsafe(name), array_index, wigname)

    def _add_label(self, ctx, cls, cls_attrs, name, elt_inst,
                                                      no_label=False, **kwargs):
        wants_no_label = cls_attrs.label is False or no_label or not self.label

        if not wants_no_label:
            elt_inst.label = self.trc(cls, ctx.locale, name)
            elt_inst.always_float_label = True

        return elt_inst

    def _gen_input_name(self, name, array_index=None):
        if array_index is None:
            return name
        return "%s[%d]" % (name, array_index)

    def _gen_widget_data(self, ctx, cls, inst, name, cls_attrs, **kwargs):
        wgt_inst_data = dict(
            label=self.trc(cls, ctx.locale, name),
            name=self._gen_input_name(name),
            type='text',
        )

        elt_class = oset([
            camel_case_to_uscore(cls.get_type_name()),
            name.rsplit(self.hier_delim, 1)[-1],
            re.sub(r'\[[0-9]+\]', '', name).replace(self.hier_delim, '__'),
        ])

        wgt_inst_data["class_"] = ' '.join(elt_class)

        if cls_attrs.pattern is not None:
            wgt_inst_data["pattern"] = cls_attrs.pattern

        if cls_attrs.read_only:
            wgt_inst_data["readonly"] = True

        if cls_attrs.error_message is not None:
            wgt_inst_data['error_message'] = \
                             self.trd(cls_attrs.error_message, ctx.locale, name)

        placeholder = cls_attrs.placeholder
        if isinstance(placeholder, six.string_types):
            wgt_inst_data["placeholder"] = placeholder

        elif placeholder:
            wgt_inst_data["placeholder"] = self.trc(cls, ctx.locale, name)

        # Required bool means, in HTML context, a checkbox that needs to be
        # checked, which is not what we mean here at all.
        if not issubclass(cls, Boolean):
            # We used OR here because html forms send empty values anyway. So a
            # missing value is sent as null as well.
            if cls_attrs.min_occurs >= 1 or cls_attrs.nullable == False:
                wgt_inst_data["required"] = True

        if (cls_attrs.write_once and inst is not None) or \
                                                       cls_attrs.write is False:
            wgt_inst_data["readonly"] = True

        if cls_attrs.min_occurs == 1 and cls_attrs.nullable == False:
            wgt_inst_data['required'] = True

        if cls_attrs.read_only:
            wgt_inst_data['readonly'] = True

        # we need to check for an explicit False and not something that walks
        # and quacks like a False.
        if cls_attrs.write is False:
            wgt_inst_data['readonly'] = True

        return wgt_inst_data

    def _gen_widget_inst(self, ctx, cls, inst, name, cls_attrs,
                                                   sugcls=PaperInput, **kwargs):

        wgt_inst_data = self._gen_widget_data(ctx, cls, inst,
                                                      name, cls_attrs, **kwargs)

        values = cls_attrs.values
        if values is None or len(values) == 0:
            wgt_inst_data['id'] = self._gen_elt_id(name, sugcls, **kwargs)
            return sugcls(**wgt_inst_data)

        sugcls = PaperDropdownMenu
        wgt_inst_data['id'] = self._gen_elt_id(name, sugcls, **kwargs)
        elt_inst = sugcls(
            listbox=PaperListbox(items=[], class_='dropdown-content'),
            valueattr="value", **wgt_inst_data)

        field_name = name.split('.')[-1]

        values_dict = cls_attrs.values_dict
        if values_dict is None:
            values_dict = {}

        inststr = self.to_unicode(cls, inst)
        if cls_attrs.write is False and inststr is not None:
            inst_label = values_dict.get(inst, inststr)
            if isinstance(inst_label, dict):
                inst_label = self.trd(inst_label, ctx.locale, field_name)

            logger.debug("\t\tinst %r label %r", inst_label, inst)
            item = PaperItem(value=inststr, label=inst_label)
            elt_inst.listbox.items.append(item)
            elt_inst.listbox.selected = 0

        else:
            ever_selected = False
            we_have_empty = False

            if cls_attrs.nullable or cls_attrs.min_occurs == 0:
                elt_inst.listbox.items.append(PaperItem(value="", label=""))
                we_have_empty = True
                # if none are selected, this will be the default anyway, so
                # no need to add selected attribute to the option tag.

            # FIXME: cache this!
            for i, v in enumerate(cls_attrs.values):
                valstr = self.to_unicode(cls, v)
                if valstr is None:
                    valstr = ""

                if inst == v:
                    ever_selected = True
                    elt_inst.listbox.selected = i

                val_label = values_dict.get(v, valstr)
                logger.debug("\t\tother values inst %r label %r", v, val_label)
                if isinstance(val_label, dict):
                    val_label = self.trd(val_label, ctx.locale, valstr)

                item = PaperItem(value=valstr, label=val_label)
                elt_inst.listbox.items.append(item)

            if not (ever_selected or we_have_empty):
                elt_inst.listbox.items.insert(0, PaperItem(value="", label=""))

        return elt_inst


class PolymerComplexDropdownWidget(PolymerWidgetBase):
    def __init__(self, text_field=None, id_field=None, data_source=None,
                                  need_parent_params=True, param_whitelist=None,
                  app=None, encoding='utf8', mime_type=None, ignore_uncap=False,
         ignore_wrappers=False, cloth=None, cloth_parser=None, polymorphic=True,
                 strip_comments=True, hier_delim='.', doctype=None, label=True):

        super(PolymerComplexDropdownWidget, self) \
            .__init__(app=app, encoding=encoding, doctype=doctype,
                      hier_delim=hier_delim, mime_type=mime_type,
                     ignore_uncap=ignore_uncap, ignore_wrappers=ignore_wrappers,
                cloth=cloth, cloth_parser=cloth_parser, polymorphic=polymorphic,
                                     strip_comments=strip_comments, label=label)

        self.serialization_handlers = cdict({
            ComplexModelBase: self.complex_model_to_parent,
        })

        self.param_whitelist = param_whitelist
        self.data_source = data_source
        self.text_field = text_field
        self.need_parent_params = need_parent_params
        self.id_field = id_field

    def complex_model_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        wgt_inst_data = self._gen_widget_data(ctx, cls, inst, name, cls_attrs,
                                                                       **kwargs)

        wgt_inst = NeuronsComplexDropdown(
            need_parent_params=self.need_parent_params,
            data_source=self.data_source,
            attr_item_label=self.text_field, attr_item_value=self.id_field,
                                                                **wgt_inst_data)

        if self.param_whitelist is not None:
            wgt_inst.param_whitelist = json.dumps(self.param_whitelist)

        self._add_label(ctx, cls, cls_attrs, name, wgt_inst, **kwargs)
        self._write_elt_inst(ctx, wgt_inst, parent)


class PolymerComplexHrefWidget(PolymerWidgetBase):
    def __init__(self, text_field=None, id_field=None,
            need_parent_params=True, param_whitelist=None, base_href=None,
                  app=None, encoding='utf8', mime_type=None, ignore_uncap=False,
         ignore_wrappers=False, cloth=None, cloth_parser=None, polymorphic=True,
                 strip_comments=True, hier_delim='.', doctype=None, label=True):

        super(PolymerComplexHrefWidget, self) \
            .__init__(app=app, encoding=encoding, doctype=doctype,
                                 hier_delim=hier_delim, mime_type=mime_type,
                     ignore_uncap=ignore_uncap, ignore_wrappers=ignore_wrappers,
                cloth=cloth, cloth_parser=cloth_parser, polymorphic=polymorphic,
                                     strip_comments=strip_comments, label=label)

        self.serialization_handlers = cdict({
            ComplexModelBase: self.complex_model_to_parent,
        })

        self.param_whitelist = param_whitelist
        self.text_field = text_field
        self.need_parent_params = need_parent_params
        self.id_field = id_field
        self.base_href = base_href

    def complex_model_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        wgt_inst_data = self._gen_widget_data(ctx, cls, inst, name, cls_attrs,
                                                                       **kwargs)
        base_href = self.base_href
        if base_href is None:
            base_href = camel_case_to_uscore(cls.get_type_name())

        wgt_inst = NeuronsComplexHref(
            base_href=base_href,
            need_parent_params=self.need_parent_params,
            attr_item_label=self.text_field, attr_item_value=self.id_field,
                                                                **wgt_inst_data)

        if self.param_whitelist is not None:
            wgt_inst.param_whitelist = json.dumps(self.param_whitelist)

        self._add_label(ctx, cls, cls_attrs, name, wgt_inst, **kwargs)
        self._write_elt_inst(ctx, wgt_inst, parent)



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import os
import hmac
import hashlib
import uuid

import msgpack
import neurons

from time import time
from Crypto.Cipher import AES

from neurons.base.error import TamperedCookieError, SessionExpiredError
from spyne.error import ValidationError
from spyne.util.six import PY3

SESSION_LIFETIME = 15 * 60


def pad(data, length):
    """
    PKCS #7-style padding with the given block length
    """
    assert length < 256
    assert length > 0
    padlen = length - len(data) % length
    assert padlen <= length
    assert padlen > 0
    return data + padlen * chr(padlen)


def unpad(data, length):
    """
    PKCS #7-style unpadding with the given block length
    """
    assert length < 256
    assert length > 0
    padlen = ord(data[-1])
    assert padlen <= length
    assert padlen > 0
    assert data[-padlen:] == padlen * chr(padlen)
    return data[:-padlen]


def encode(data, secret, salt=None):
    """Encode and return the data as a random-IV-prefixed AES-encrypted
    HMAC-SHA1-authenticated padded message corresponding to the given
    data string and secret, which should be at least 36 randomly
    chosen bytes agreed upon by the encoding and decoding parties.
    """

    assert len(secret) >= 36
    if salt is None:
        salt = os.urandom(16)

    aes = AES.new(secret[:16], AES.MODE_CBC, salt)
    padded_data = pad(20 * '\0' + data, 16)[20:]
    mac = hmac.new(key=secret[16:], msg=padded_data, digestmod=hashlib.sha1).digest()
    encrypted = aes.encrypt(mac + padded_data)

    return salt + encrypted


def decode(data, secret):
    """Decode and return the data from random-IV-prefixed AES-encrypted
    HMAC-SHA1-authenticated padded message corresponding to the given
    data string and secret, which should be at least 36 randomly
    chosen bytes agreed upon by the encoding and decoding parties.
    """

    assert len(secret) >= 36

    if len(data) < 16:
        raise ValidationError(data, "Invalid session value. Please sign out "
                            "and delete cookies and try again.\nData: '%s'")
    salt = data[:16]

    encrypted = data[16:]
    aes = AES.new(secret[:16], AES.MODE_CBC, salt)
    decrypted = aes.decrypt(encrypted)

    mac = decrypted[:20]
    padded_data = decrypted[20:]
    mac2 = hmac.new(key=secret[16:], msg=padded_data, digestmod=hashlib.sha1).digest()

    if mac != mac2:
        raise TamperedCookieError()

    return unpad(20 * '\0' + padded_data, 16)[20:]


class SessionObject(object):
    def __init__(self, shash, user, domain, ttl):
        """
        :param shash: Session hash: Unique session identifier.
        :param user: Kullanıcı adı.
        :param domain: Alan adı.
        :param ttl: Time-to-live -- eldeki session stringinin geçerlilik
               süresi.
        """

        object.__init__(self)
        self.shash = shash
        self.user = user
        self.domain = domain
        self.ttl = ttl


def get_du_from_sid(sid):
    """Session-ID stringini decrypt ederek içindeki domain_name / user_name
    bilgisini döndürür."""

    session = get_session_data(sid)
    return session.domain, session.user


def get_data(data, cls):
    dec = decode(''.join(data), neurons.secret)
    dec_data = msgpack.loads(dec)
    dec_data = cls(*dec_data)

    if dec_data.ttl <= time():
        raise SessionExpiredError()
    else:
        return dec_data


def get_session_data(sid):
    return get_data(sid, SessionObject)


def hashpass(password):
    salt = os.urandom(32)
    s = hashlib.sha1()
    s.update(password)
    s.update(salt)
    hash = s.digest() + salt
    phash = '{SSHA}' + hash

    return phash


def replenish_sid(sid):
    s = get_session_data(sid)

    return replenish_session(s.domain, s.user, s.shash)[0]


def replenish_sid_ttl(sid):
    s = get_session_data(sid)

    return replenish_session(s.domain, s.user, s.shash)


def put_data(seconds, *args):
    ttl = time() + seconds
    data = msgpack.dumps( args + (ttl,) )
    enc = encode(data, neurons.secret)

    return enc, ttl


def replenish_session(domain, user, session_id):
    return put_data(SESSION_LIFETIME, session_id, user, domain)


def gen_session_hash(user):
    salt = os.urandom(16)
    if PY3:
        key = b''.join([uuid.getnode().to_bytes(), user, time().to_bytes()])
    else:
        key = b''.join([str(uuid.getnode()), user, str(time())])

    s = hashlib.sha1()
    s.update(key)
    s.update(salt)

    return s.digest()


def verify_hash(clear_password, password_hash):
    if password_hash.startswith("{SSHA}"):
        digest = password_hash[6:26]
        salt = password_hash[26:]

        hr = hashlib.sha1(clear_password.encode('utf8'))
        hr.update(salt)

        return digest == hr.digest()

    raise NotImplementedError(password_hash[:password_hash.find('}')+1])



#!/usr/bin/env python
# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import absolute_import, print_function

import logging
logging.basicConfig(level=logging.DEBUG)

import unittest

from lxml import html, etree

from neurons.polymer import PolymerForm

from spyne import Application, NullServer, Service, rpc, Unicode, Integer, \
    ComplexModel, Decimal
from spyne.util.test import show


def _test_type(cls, inst):
    # silence bogus warnings
    from spyne.util import appreg; appreg.applications.clear()

    class SomeService(Service):
        @rpc(_returns=cls, _body_style='bare')
        def some_call(ctx):
            return inst

    prot = PolymerForm()
    app = Application([SomeService], 'some_ns', out_protocol=prot)

    null = NullServer(app, ostr=True)

    ret = ''.join(null.service.some_call())
    try:
        elt = html.fromstring(ret)
    except:
        print(ret)
        raise

    show(elt, stdout=False)
    print(etree.tostring(elt, pretty_print=True))

    return elt


class TestPolymerForm(object):
    def test_simple_decimal(self):
        v = 42
        elt = _test_type(Decimal(ge=0, le=5), v)

        assert elt.xpath('paper-input/@name') == ['']
        assert elt.xpath('paper-input/@type') == ['number']
        assert elt.xpath('paper-input/@min') == ['0']
        assert elt.xpath('paper-input/@max') == ['5']

    def test_complex(self):
        class SomeObject(ComplexModel):
            _type_info = [
                ('i', Integer),
                ('s', Unicode),
            ]

        v = SomeObject(i=42, s="Arthur")
        elt = _test_type(SomeObject, v)

        assert elt.xpath('fieldset/paper-input/@name') == ['i', 's']


class TestPolymerDropdownMenu(object):
    def test_simple(self):
        elt = _test_type(Integer(values=[0, 3, 2, 1]), None)
        ret = elt.xpath('paper-dropdown-menu/paper-listbox/paper-item/text()')

        # This has to do with the workaround in complex_to_parent in
        # cloth/to_parent.py
        ret = [r.strip() for r in ret]

        assert ret == ['', '0', '3', '2', '1']

    def _test_complex(self):
        class SomeObject(ComplexModel):
            class Attributes(ComplexModel.Attributes):
                prot = PolymerDropdownMenu('s', 'i')

            _type_info = [
                ('i', Integer),
                ('s', Unicode),
            ]

        v = SomeObject(i=42, s="Arthur")
        elt = _test_type(SomeObject, v)

        assert elt.xpath('paper-dropdown-menu/paper-listbox/paper-item/@name') \
                                                                         == ['']


if __name__ == '__main__':
    unittest.main()



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import logging
logger = logging.getLogger(__name__)

import threading
import traceback

import neurons

from spyne.util.color import G, YEL, R

from sqlalchemy import MetaData
from sqlalchemy.orm import sessionmaker
from sqlalchemy.engine import Engine


try:
    import ldap
except ImportError as e:
    class _catchall(object):
        def __getattribute__(self, item):
            raise e

    ldap = _catchall()

    del _catchall


class DataStoreBase(object):
    def __init__(self, name, type):
        self.name = name
        self.type = type


class LdapDataStore(DataStoreBase):
    SUPPORTED_BACKENDS = ('python-ldap', )  # TODO: add ldaptor for python3

    def __init__(self, name, parent, type='python-ldap'):
        assert type in LdapDataStore.SUPPORTED_BACKENDS
        DataStoreBase.__init__(self, name, type)

        self.conn = None
        self.parent = parent

    def apply_simple(self, bind_dn=None, password=None):
        parent = self.parent

        if bind_dn is None:
            bind_dn = parent.bind_dn

        if password is None:
            password = parent.password

        if parent.use_tls:
            uri = "ldaps://%s:%d" % (parent.host, parent.port)
            retval = ldap.initialize(uri)

            retval.set_option(ldap.OPT_X_TLS, ldap.OPT_X_TLS_DEMAND)
            retval.set_option(ldap.OPT_X_TLS_DEMAND, True)

        else:
            uri = "ldap://%s:%d" % (parent.host, parent.port)
            retval = ldap.initialize(uri)

        retval.set_option(ldap.OPT_NETWORK_TIMEOUT, parent.timeout)

        if not parent.referrals:
            retval.set_option(ldap.OPT_REFERRALS, 0)

        if parent.version == 3:
            retval.protocol_version = ldap.VERSION3

        elif parent.version == 2:
            retval.protocol_version = ldap.VERSION2

        else:
            raise ValueError(parent.version)

        retval.simple_bind_s(bind_dn, password)

        logger.info("Ldap connection to %s successful.",
                                                retval.get_option(ldap.OPT_URI))

        return retval

    def apply(self):
        if self.conn is not None:
            self.conn.close()

        if self.parent.method == 'simple':
            self.conn = self.apply_simple()

        elif self.parent.method == 'sasl':
            raise NotImplementedError(self.parent.method)

        else:
            raise ValueError(self.parent.method)


# FIXME: get rid of the overly complicated property setters.
class SqlDataStore(DataStoreBase):
    def __init__(self, name=None, connection_string=None,
                                          engine=None, metadata=None, **kwargs):
        DataStoreBase.__init__(self, name=name, type='sqlalchemy')

        if engine is not None:
            assert isinstance(engine, Engine)

        if metadata is not None:
            assert isinstance(metadata, MetaData)

        self.__kwargs = kwargs
        self.__metadata = None
        self.__engine = None
        self.Session = None
        """SQLAlchemy session constructor."""

        self.metadata = metadata or MetaData()
        self.engine = engine
        self.Session = sessionmaker()
        self.connection_string = connection_string

        self.txpool = None
        """TxPostgres connection pool. Added when `add_txpool` is called."""

        self.txpool_min = 1
        """TxPostgres minimum number of pooled connections."""

        self.txpool_start_deferred = None
        """Deferred from TxPostgres pool start()."""

    @property
    def txpool(self):
        if neurons.REACTOR_THREAD_ID is not None and \
                  neurons.REACTOR_THREAD_ID != threading.current_thread().ident:

            logger.warning("Using txpostgres outside of reactor thread is "
                           "dangerous.")

            for elt in traceback.format_stack():
                for line in elt.split("\n"):
                    if len(line) > 0:
                        logger.warning(line)

        return self._txpool

    @txpool.setter
    def txpool(self, what):
        self._txpool = what

    def add_txpool(self):
        # don't import twisted too soon
        from txpostgres.txpostgres import Connection, ConnectionPool
        from txpostgres.reconnection import DeadConnectionDetector

        class LoggingDeadConnectionDetector(DeadConnectionDetector):
            NAME_G = G('{%s}' % (self.name,))
            NAME_R = R('{%s}' % (self.name,))
            NAME_YEL = YEL('{%s}' % (self.name,))

            def startReconnecting(self, err):
                logger.warning('%s (txpool) database connection down: %r)',
                                                         self.NAME_R, err.value)
                return DeadConnectionDetector.startReconnecting(self, err)

            def reconnect(self):
                logger.warning('%s (txpool) reconnecting...', self.NAME_YEL)
                return DeadConnectionDetector.reconnect(self)

            def connectionRecovered(self):
                logger.warning('%s (txpool) connection recovered.', self.NAME_G)
                return DeadConnectionDetector.connectionRecovered(self)

        try:
            dsn = self.engine.raw_connection().connection.dsn
        except Exception:
            print("Error getting dsn for conn_str", self.connection_string)
            raise

        txpool_min = self.txpool_min

        class NeuronsConnectionPool(ConnectionPool):
            @staticmethod
            def connectionFactory(reactor=None, cooperator=None):
                retval = Connection(reactor=reactor, cooperator=cooperator,
                                       detector=LoggingDeadConnectionDetector())

                logger.debug("{%s} (txpool) spawning backend", self.name)
                return retval

            def __repr__(self):
                data = (
                    ', '.join(repr(c) for c in self.connargs),
                    'min=%d' % (txpool_min,),
                    ', '.join(("%s=%r" % (k, v)
                                              for k, v in self.connkw.items())),
                )

                data = [s for s in data if len(s) > 0]

                return "NeuronsConnectionPool(%s)" % (', '.join(data),)

        self.txpool = NeuronsConnectionPool("heleleley", dsn,
                                                            min=self.txpool_min)
        self.txpool_start_deferred = self.txpool.start()
        self.txpool_start_deferred \
            .addCallback(lambda p:
                logger.info("{%s} (txpool) %r started with dsn: '%s'.",
                                                           self.name, p, dsn)) \
            .addErrback(lambda err: err.printTraceback())

        return self.txpool_start_deferred

    def connect(self):
        return self.__engine.connect()

    @property
    def meta(self):
        return self.__metadata

    @property
    def metadata(self):
        return self.__metadata

    @metadata.setter
    def metadata(self, metadata):
        self.__metadata = metadata
        if self.__engine is not None and metadata is not None:
            self.__metadata.bind = self.__engine

    @property
    def engine(self):
        return self.__engine

    @engine.setter
    def engine(self, engine):
        self.__engine = engine
        if engine is not None:
            if self.metadata is not None:
                self.metadata.bind = engine
            if self.Session is not None:
                self.Session.configure(bind=engine, expire_on_commit=False)

    @property
    def kwargs(self):
        return self.__kwargs

    @property
    def connection_string(self):
        return self.__connection_string

    @connection_string.setter
    def connection_string(self, what):
        from sqlalchemy.engine import create_engine
        self.__connection_string = what

        if what is None:
            self.engine = None

        else:
            # FIXME: parse the url properly! it's not that difficult!
            if what.startswith('sqlite://') or what.endswith(":memory:"):
                from sqlalchemy.pool import StaticPool

                self.__kwargs['poolclass'] = StaticPool

                if what.startswith('sqlite'):
                    self.__kwargs['connect_args'] = {'check_same_thread': False}
                if 'pool_size' in self.__kwargs:
                    del self.__kwargs['pool_size']

            self.engine = create_engine(what, **self.__kwargs)
            try:
                dsn = self.engine.raw_connection().connection.dsn
            except Exception:
                dsn = None

            logger.info("{%s} (sqla) %r started with args: %r dsn: %r",
                                       self.name, self.engine, self.kwargs, dsn)


def get_data_store(type, *args, **kwargs):
    if type == 'ldap':
        return LdapDataStore(*args, **kwargs)
    elif type == 'sqlalchemy':
        return SqlDataStore(*args, **kwargs)
    else:
        raise ValueError("Unrecognized data store %r" % type)



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logger = logging.getLogger(__name__)

from os.path import abspath

from spyne import Application
from spyne.protocol.html import HtmlMicroFormat
from spyne.protocol.http import HttpRpc

from twisted.internet import reactor

from neurons.daemon.config import HttpServer, StaticFileServer

from garage.const import T_INDEX
from garage.service import GarageService


def start_garage(config):
    subconfig = config.services.getwrite('http', HttpServer(
        host='0.0.0.0',
        port=7111,
        disabled=False,
        _subapps=[StaticFileServer(url='assets', path=abspath('assets'),
                                                          list_contents=False)],
    ))

    subconfig.subapps[''] = \
        Application([GarageService], 'garage.main',
                in_protocol=HttpRpc(validator='soft'),
                out_protocol=HtmlMicroFormat(cloth=T_INDEX)
            )

    logger.info("listening for garage http on %s:%d",
                                                 subconfig.host, subconfig.port)

    return reactor.listenTCP(subconfig.port, subconfig.gen_site(),
                                                 interface=subconfig.host), None



# encoding: utf8
#
# retrieved from http://www.aminus.net/wiki/Dowser at 2015-03-18
# this document were placed in public domain by their author
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

import logging
logger = logging.getLogger(__name__)

import cgi
import os
import gc
import sys

from types import FrameType, ModuleType
from collections import defaultdict

from neurons.daemon.dowser.const import ASSETS_DIR

from spyne import rpc, Unicode, Service, ByteArray, AnyHtml
from spyne import Integer
from spyne.protocol.http import HttpPattern, HttpRpc
from spyne.util.six import BytesIO

from neurons.daemon.dowser import reftree


try:
    from PIL import Image
    from PIL import ImageDraw

except ImportError as e:
    class Image:
        @classmethod
        def new(self):
            raise e

    ImageDraw = None


def get_repr(obj, limit=250):
    return cgi.escape(reftree.get_repr(obj, limit))


class _(object): pass


dictproxy = type(_.__dict__)

method_types = [
    type(tuple.__le__),  # 'wrapper_descriptor'
    type([1].__le__),  # 'method-wrapper'
    type(sys.getcheckinterval),  # 'builtin_function_or_method'
    type(cgi.FieldStorage.getfirst),  # 'instancemethod'
]


def template(name, **params):
    p = {
        'maincss': "/assets/main.css",
        'home': "/",
    }
    p.update(params)
    return open(os.path.join(ASSETS_DIR, name)).read() % p


class DowserServices(Service):
    period = 5
    maxhistory = 300
    history = {}
    samples = 0
    id_history = []
    max_id_history = 2

    @classmethod
    def tick(cls):
        logger.debug("Dowser tick")
        gc.collect()

        typecounts = {}
        new_ids = defaultdict(set)
        for obj in gc.get_objects():
            objtype = type(obj)
            typename = ".".join((objtype.__module__, objtype.__name__))
            new_ids[typename].add(id(obj))

            if objtype in typecounts:
                typecounts[objtype] += 1
            else:
                typecounts[objtype] = 1

        if len(DowserServices.id_history) > DowserServices.max_id_history:
            DowserServices.id_history.pop(0)
        DowserServices.id_history.append(new_ids)

        for objtype, count in typecounts.items():
            typename = objtype.__module__ + "." + objtype.__name__
            if typename not in cls.history:
                cls.history[typename] = [0] * cls.samples
            cls.history[typename].append(count)

        samples = cls.samples + 1

        # Add dummy entries for any types which no longer exist
        for typename, hist in cls.history.items():
            diff = samples - len(hist)
            if diff > 0:
                hist.extend([0] * diff)

        # Truncate history to self.maxhistory
        if samples > cls.maxhistory:
            for typename, hist in cls.history.items():
                hist.pop(0)
        else:
            cls.samples = samples

    ORDER_MAP = {
        None: None,
        'min': lambda x: min(DowserServices.history[x]),
        'cur': lambda x: DowserServices.history[x][-1],
        'max': lambda x: max(DowserServices.history[x]),
        'diff': lambda x: len(DowserServices.id_history[-1][x]) - \
                          len(DowserServices.id_history[0][x]),
    }

    @rpc(Unicode(values=['max', 'min', 'cur', 'diff']), Integer(default=0),
                     Integer(default=0), Integer(default=0), Integer(default=0),
                     _patterns=[HttpPattern('/', verb='GET')], _returns=AnyHtml)
    def index(ctx, order, cur_floor, min_floor, max_floor, diff_floor):
        rows = []
        typenames = ctx.descriptor.service_class.history.keys()

        key = ctx.descriptor.service_class.ORDER_MAP[order]
        typenames.sort(key=key, reverse=False if order is None else True)

        for typename in typenames:
            hist = ctx.descriptor.service_class.history[typename]

            # get numbers
            maxhist = max(hist)
            minhist = max(hist)
            cur = hist[-1]

            idhist = DowserServices.id_history
            last = idhist[-1][typename]
            first = idhist[0][typename]
            diff = len(last) - len(first)

            # check floors
            show_this = cur >= cur_floor and \
                        minhist >= min_floor and \
                        maxhist >= max_floor and \
                        abs(diff) >= diff_floor

            if not show_this:
                continue

            row = (
                '<div class="typecount">%s<br />'
                '<img class="chart" src="%s" /><br />'
                'Min: %s Cur: %s Max: %s Diff: %s '
                '<a href="%s">TRACE</a> &middot; '
                '<a href="%s">DIFF TRACE</a>'
                '</div>' % (
                    cgi.escape(typename),
                    "/chart?typename=%s" % typename,
                    minhist, cur, maxhist, diff,
                    "/trace?typename=%s" % typename,
                    "/difftrace?typename=%s" % typename,
                )
            )
            rows.append(row)

        return template("graphs.html", output="\n".join(rows))

    @rpc(Unicode, _returns=ByteArray)
    def chart(ctx, typename):
        """Return a sparkline chart of the given type."""

        data = ctx.descriptor.service_class.history[typename]
        height = 20.0
        scale = height / max(data)
        im = Image.new("RGB", (len(data), int(height)), 'white')
        draw = ImageDraw.Draw(im)
        draw.line([(i, int(height - (v * scale))) for i, v in enumerate(data)],
                                                                 fill="#009900")
        del draw

        f = BytesIO()
        im.save(f, "PNG")
        result = f.getvalue()

        ctx.out_protocol = HttpRpc()
        ctx.transport.resp_headers["Content-Type"] = "image/png"
        return [result]

    @rpc(Unicode, Integer, _returns=AnyHtml)
    def trace(ctx, typename, objid):
        gc.collect()

        if objid is None:
            rows = DowserServices.trace_all(typename)
        else:
            rows = DowserServices.trace_one(typename, objid)

        return template("trace.html", output="\n".join(rows),
                        typename=cgi.escape(typename),
                        objid=str(objid or ''))

    @rpc(Unicode, _returns=AnyHtml)
    def difftrace(ctx, typename):
        gc.collect()

        rows = DowserServices.trace_all(typename, difftrace=True)

        return template("trace.html", output="\n".join(rows),
                                        typename=cgi.escape(typename), objid='')

    @classmethod
    def trace_all(cls, typename, difftrace=False):
        rows = []
        for obj in gc.get_objects():
            objtype = type(obj)
            if objtype.__module__ + "." + objtype.__name__ == typename and \
                    ((not difftrace) or (
                             id(obj) in DowserServices.id_history[-1] and
                        (not id(obj) in DowserServices.id_history[0])
                    )):
                rows.append("<p class='obj'>%s</p>"
                            % ReferrerTree(obj).get_repr(obj))

        if not rows:
            rows = ["<h3>The type you requested was not found.</h3>"]

        return rows

    @classmethod
    def trace_one(cls, typename, objid):
        rows = []
        all_objs = gc.get_objects()
        for obj in all_objs:
            if id(obj) == objid:
                objtype = type(obj)
                if objtype.__module__ + "." + objtype.__name__ != typename:
                    rows = ["<h3>The object you requested is no longer "
                            "of the correct type.</h3>"]
                else:
                    # Attributes
                    rows.append('<div class="obj"><h3>Attributes</h3>')
                    for k in dir(obj):
                        v = getattr(obj, k)
                        if type(v) not in method_types:
                            rows.append('<p class="attr"><b>%s:</b> %s</p>' %
                                        (k, get_repr(v)))
                        del v
                    rows.append('</div>')

                    # Referrers
                    rows.append(
                        '<div class="refs"><h3>Referrers (Parents)</h3>')
                    rows.append('<p class="desc"><a href="%s">Show the '
                                'entire tree</a> of reachable objects</p>'
                                % (
                    "/tree?typename=%s&objid=%s" % (typename, objid)))
                    tree = ReferrerTree(obj)
                    tree.ignore(all_objs)
                    for depth, parentid, parentrepr in tree.walk(maxdepth=1):
                        if parentid:
                            rows.append("<p class='obj'>%s</p>" % parentrepr)
                    rows.append('</div>')

                    # Referents
                    rows.append(
                        '<div class="refs"><h3>Referents (Children)</h3>')
                    for child in gc.get_referents(obj):
                        rows.append(
                            "<p class='obj'>%s</p>" % tree.get_repr(child))
                    rows.append('</div>')
                break
        if not rows:
            rows = ["<h3>The object you requested was not found.</h3>"]
        return rows

    @rpc(Unicode, Integer, _returns=AnyHtml)
    def tree(self, typename, objid):
        gc.collect()

        rows = []
        all_objs = gc.get_objects()
        for obj in all_objs:
            if id(obj) == objid:
                objtype = type(obj)
                if objtype.__module__ + "." + objtype.__name__ != typename:
                    rows = ["<h3>The object you requested is no longer "
                            "of the correct type.</h3>"]
                else:
                    rows.append('<div class="obj">')

                    tree = ReferrerTree(obj)
                    tree.ignore(all_objs)
                    for depth, parentid, parentrepr in tree.walk(
                            maxresults=1000):
                        rows.append(parentrepr)

                    rows.append('</div>')
                break

        if not rows:
            rows = ["<h3>The object you requested was not found.</h3>"]

        params = {
            'output': "\n".join(rows),
            'typename': cgi.escape(typename),
            'objid': str(objid),
        }

        return template("tree.html", **params)


class ReferrerTree(reftree.Tree):
    ignore_modules = True

    def _gen(self, obj, depth=0):
        if self.maxdepth and depth >= self.maxdepth:
            yield depth, 0, "---- Max depth reached ----"
            raise StopIteration()

        if isinstance(obj, ModuleType) and self.ignore_modules:
            raise StopIteration()

        refs = gc.get_referrers(obj)
        refiter = iter(refs)
        self.ignore(refs, refiter)
        thisfile = sys._getframe().f_code.co_filename
        for ref in refiter:
            # Exclude all frames that are from this module or reftree.
            if (isinstance(ref, FrameType)
                       and ref.f_code.co_filename in (thisfile, self.filename)):
                continue

            # Exclude all functions and classes from this module or reftree.
            mod = getattr(ref, "__module__", "")
            if mod is None:
                continue

            if "dowser" in mod or "reftree" in mod or mod == '__main__':
                continue

            # Exclude all parents in our ignore list.
            if id(ref) in self._ignore:
                continue

            # Yield the (depth, id, repr) of our object.
            yield depth, 0, '%s<div class="branch">' % (" " * depth)

            if id(ref) in self.seen:
                yield depth, id(ref), "see %s above" % id(ref)

            else:
                self.seen[id(ref)] = None
                yield depth, id(ref), self.get_repr(ref, obj)

                for parent in self._gen(ref, depth + 1):
                    yield parent

            yield depth, 0, '%s</div>' % (" " * depth)

    def get_repr(self, obj, referent=None):
        """Return an HTML tree block describing the given object."""
        objtype = type(obj)
        typename = objtype.__module__ + "." + objtype.__name__
        prettytype = typename.replace("__builtin__.", "")

        name = getattr(obj, "__name__", "")
        if name:
            prettytype = "%s %r" % (prettytype, name)

        key = ""
        if referent:
            key = self.get_refkey(obj, referent)
        return ('<a class="objectid" href="%s">%s</a> '
                '<span class="typename">%s</span>%s<br />'
                '<span class="repr">%s</span>'
                % (("/trace?typename=%s&objid=%s" % (typename, id(obj))),
                   id(obj), prettytype, key, get_repr(obj, 100))
                )

    def get_refkey(self, obj, referent):
        """Return the dict key or attribute name of obj which refers to
        referent."""
        if isinstance(obj, dict):
            for k, v in obj.items():
                if v is referent:
                    return " (via its %r key)" % (k,)

        for k in dir(obj) + ['__dict__']:
            if getattr(obj, k, None) is referent:
                return " (via its %r attribute)" % (k,)
        return ""



#!/usr/bin/env python
# encoding: utf8
#
# Copyright © Burak Arslan <burak at arskom dot com dot tr>,
#             Arskom Ltd. http://www.arskom.com.tr
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#    1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#    2. Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#    3. Neither the name of the owner nor the names of its contributors may be
#       used to endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER BE LIABLE FOR ANY DIRECT,
# INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
# OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
# EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import absolute_import

from spyne import Unicode, Integer, Array, Integer32, table

from neurons.model import TableModel

p = lambda s: "neurons_%s" % s


class Vehicle(TableModel):
    __tablename__ = p('vehicle')

    __mapper_args__ = {
        'polymorphic_on': 'type',
        'polymorphic_identity': 'base',
        'with_polymorphic': '*',  # Bir query'de tum fieldlari ceksin diye.
    }

    _type_info = [
        ('id', Integer32(pk=True)),
        ('owner', Unicode),
        ('type', Unicode(8)),
    ]


class Car(Vehicle):
    _type_info = [
        ('color', Unicode),
        ('speed', Integer),
    ]


class Bike(Vehicle):
    _type_info = [
        ('size', Integer),
    ]


class Garage(TableModel):
    __tablename__ = p('garage')

    _type_info = [
        ('id', Integer32(pk=True)),
        ('vehicles', Array(Vehicle, store_as=table(right='garage_id'))),
    ]



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import print_function

import logging
logger = logging.getLogger(__name__)

from collections import namedtuple
from inspect import isgenerator
from decimal import Decimal as D

from lxml import etree, html
from lxml.builder import E

from spyne import ComplexModelBase, Unicode, Decimal, Boolean, Date, Time, \
    DateTime, Integer, Duration, PushBase, Array, Uuid, AnyHtml, AnyXml, \
    AnyUri, Fault, File
from spyne.util import coroutine, Break, six, memoize_id
from spyne.util.cdict import cdict
from spyne.server.http import HttpTransportContext
try:
    from spyne.store.relational.document import FileData
except ImportError:
    FileData = type('__ignored', (object,), {})

from neurons.form.widget import HtmlFormWidget, SimpleRenderWidget

SOME_COUNTER = 0


class TabFsetState(object):
    def __init__(self):
        self.prev_fset = None
        self.fset_ctx = None
        self.fset = None

        self.prev_tab = None
        self.tab_ctx = None
        self.tab = None

        self.tabview_ctx = None
        self.tabview_id = None


def _idxof(haystack, needle, default):
    try:
        return haystack.index(needle)
    except ValueError:
        return default


class Fieldset(namedtuple('Fieldset', 'legend tag attrib index htmlid')):
    def __new__(cls, legend=None, tag='fieldset', attrib={}, index=None,
                                                                   htmlid=None):
        global SOME_COUNTER
        if htmlid is None:
            htmlid = 'fset' + str(SOME_COUNTER)
            SOME_COUNTER += 1

        return super(Fieldset, cls).__new__(cls, legend, tag, attrib, index,
                                                                         htmlid)

class Tab(namedtuple('Tab', 'legend attrib index htmlid')):
    def __new__(cls, legend=None, attrib={}, index=None, htmlid=None):
        global SOME_COUNTER
        if htmlid is None:
            htmlid = "tab" + str(SOME_COUNTER)
            SOME_COUNTER += 1

        return super(Tab, cls).__new__(cls, legend, attrib, index, htmlid)


def _gen_array_js(parent, key):
    parent.write(E.script("""
if (window.$) $(function() {
    var field_name = "." + "%(field_name)s";

    var add = function() {
        var f = $($(field_name)[0]).clone();
        f.appendTo($(field_name).parent());
        f.find(".%(field_name)s_btn_add").click(add);
        f.find(".%(field_name)s_btn_del").click(del);
        return false;
    };

    var del = function(event) {
        if($('#%(field_name)s_container').find('.%(field_name)s').length > 1){
            $(this).parent().remove();
        }
        else{
            $(this).parent().children().val("")
        }
       event.preventDefault();
    };

    $(".%(field_name)s_btn_add").click(add);
    $(".%(field_name)s_btn_del").click(del);
});""" % {"field_name": key}, type="text/javascript"))


_jstag = lambda src: E.script(src=src, type="text/javascript")
_csstag = lambda src: E.link(href=src, type="text/css", rel="stylesheet")



def THtmlFormRoot(Base):
    class HtmlFormRoot(Base):
        def __init__(self, app=None, encoding='utf8',
                                      ignore_uncap=False, ignore_wrappers=False,
                cloth=None, cloth_parser=None, polymorphic=True, hier_delim='.',
                                     label=True, doctype=None, placeholder=None,
               input_class=None, input_div_class=None, input_wrapper_class=None,
                 label_class=None, action=None, method='POST', before_form=None,
                                                            mrpc_url_prefix=''):

            super(HtmlFormRoot, self).__init__(app=app, encoding=encoding,
                doctype=doctype,
                     ignore_uncap=ignore_uncap, ignore_wrappers=ignore_wrappers,
                cloth=cloth, cloth_parser=cloth_parser, polymorphic=polymorphic,
                    hier_delim=hier_delim, label=label,
                    placeholder=placeholder, input_class=input_class,
                    input_div_class=input_div_class,
               input_wrapper_class=input_wrapper_class, label_class=label_class)

            self.action = action
            self.method = method
            if before_form is not None:
                self.event_manager.add_listener("before_form", before_form)

            self.simple = SimpleRenderWidget(label=label)
            self.mrpc_url_prefix = mrpc_url_prefix

        @coroutine
        def start_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
            # FIXME: what a HUGE swath of copy/paste! I want yield from!

            # if we are already in a form, don't open a <form> tag
            if not getattr(ctx.outprot_ctx, 'in_form', False):
                ctx.outprot_ctx.in_form = True

                if not (len(ctx.outprot_ctx.prot_stack) == 1 and
                              isinstance(ctx.protocol.prot_stack[0], HtmlForm)):
                    name = ''

                attrib = self._gen_form_attrib(ctx, cls)
                self.event_manager.fire_event("before_form", ctx, cls, inst,
                                          parent, name, attrib=attrib, **kwargs)

                with parent.element(self.HTML_FORM, attrib=attrib):
                    ret = super(HtmlFormRoot, self).start_to_parent(ctx, cls,
                                                   inst, parent, name, **kwargs)
                    if isgenerator(ret):
                        try:
                            while True:
                                y = (yield)
                                ret.send(y)

                        except Break as b:
                            try:
                                ret.throw(b)
                            except StopIteration:
                                pass

                        finally:
                            self._write_submit_button(parent)
                            ctx.protocol.in_form = False
                    else:
                        self._write_submit_button(parent)
                        ctx.protocol.in_form = False

            else:
                ret = super(HtmlFormRoot, self).start_to_parent(ctx, cls, inst,
                                                         parent, name, **kwargs)
                if isgenerator(ret):
                    try:
                        while True:
                            y = (yield)
                            ret.send(y)

                    except Break as b:
                        try:
                            ret.throw(b)
                        except StopIteration:
                            pass

        def _write_submit_button(self, parent):
            parent.write(
                E.p(
                    E.input(
                        value="Submit",
                        type="submit",
                        **{'class': "btn btn-default"}
                    ),
                    **{'class': "text-center"}
                )
            )

        def _gen_form_attrib(self, ctx, cls):
            cls_attrs = self.get_cls_attrs(cls)

            attrib = dict(method=self.method)
            if self.method == 'POST':
                attrib['enctype'] = "multipart/form-data"

            attrib.update(self._get_form_action(ctx, cls_attrs))

            return attrib

        def _get_form_action(self, ctx, cls_attrs):
            if hasattr(ctx.protocol, 'form_action'):
                action = ctx.protocol.form_action
                logger.debug("Set form action to '%s' from ctx", action)
                return {'action': action}

            elif cls_attrs.form_action:
                action = cls_attrs.form_action
                logger.debug("Set form action to '%s' from cls_attrs", action)
                return {'action': action}

            elif self.action is not None:
                action = self.action
                logger.debug("Set form action to '%s' from protocol", action)
                return {'action': action}

            elif isinstance(ctx.transport, HttpTransportContext):
                action = ctx.transport.get_path()
                logger.debug("Set form action to '%s' from request path",
                                                                         action)
                return {'action': action}

            return {}

        def _check_simple(self, f):
            def _ch(ctx, cls, inst, parent, name, **kwargs):
                cls_attrs = self.get_cls_attrs(cls)
                if cls_attrs.hidden:
                    self._gen_input_hidden(cls, inst, parent, name)
                elif cls_attrs.read_only:
                    self.simple.to_parent(ctx, cls, inst, parent, name,
                                                                       **kwargs)
                else:
                    f(ctx, cls, inst, parent, name, **kwargs)

            _ch.__name__ = f.__name__
            return _ch

        def _gen_tab_headers(self, ctx, fti):
            retval = E.ul()

            tabs = {}
            for k, v in fti:
                subattr = self.get_cls_attrs(v)
                tab = subattr.tab
                if tab is not None and not subattr.exc:
                    tabs[id(tab)] = tab

            for i, tab in enumerate(sorted(tabs.values(),
                                            key=lambda t: (t.index, t.htmlid))):
                retval.append(E.li(E.a(
                    self.trd(tab.legend, ctx.locale, "Tab %d" % i),
                    href="#" + tab.htmlid
                )))

            return retval

        def _form_key(self, sort_key):
            k, v = sort_key
            attrs = self.get_cls_attrs(v)
            return None if attrs.tab is None else \
                                (attrs.tab.index, attrs.tab.htmlid), \
                   None if attrs.fieldset is None else \
                                (attrs.fieldset.index, attrs.fieldset.htmlid),

        @coroutine
        def _render_complex(self, ctx, cls, inst, parent, name, in_fset,
                                                                      **kwargs):
            global SOME_COUNTER

            fti = self.sort_fields(cls)
            fti.sort(key=self._form_key)

            s = TabFsetState()

            # FIXME: hack! why do we have top-level object receiving name?
            if name == cls.get_type_name():
                name = ''

            if in_fset:
                default_text = cls.get_type_name()
                tr = self.trc(cls, ctx.locale, default_text)
                parent.write(E.legend(tr))

            for k, v in fti:
                subattr = self.get_cls_attrs(v)
                if subattr.exc:
                    logger.debug("Excluding %s", k)
                    continue

                logger.debug("Generating form element for %s", k)
                try:
                    subinst = getattr(inst, k, None)
                except (KeyError, AttributeError):
                    subinst = None

                s.tab = subattr.tab
                if not (s.tab is s.prev_tab):
                    self._change_tab(ctx, parent, fti, s)

                s.fset = subattr.fieldset
                if not (s.fset is s.prev_fset):
                    self._change_fset(ctx, parent, k, s)

                subname = subattr.sub_name
                if subname is None:
                    subname = k

                if name is not None and len(name) > 0:
                    subkey = self.hier_delim.join((name, subname))
                else:
                    subkey = k

                label_ctxs = []
                if subattr.label:
                    if self.input_wrapper_class is not None:
                        div_attrib = {'class': self.input_wrapper_class}

                        div_ctx = parent.element('div', div_attrib)
                        div_ctx.__enter__()
                        label_ctxs.append(div_ctx)
                        logger.debug("entering label wrapper")

                    label_attrib = {}
                    if self.label_class is not None:
                        label_attrib['class'] = ' '.join((
                            self.label_class, self.selsafe(name)))

                    with parent.element('label', label_attrib):
                        parent.write(self.trc(v, ctx.locale, subkey))

                ret = self.to_parent(ctx, v, subinst, parent, subkey, **kwargs)
                if isgenerator(ret):
                    try:
                        while True:
                            y = (yield)
                            ret.send(y)

                    except Break as b:
                        try:
                            ret.throw(b)
                        except StopIteration:
                            pass

                for c in reversed(label_ctxs):
                    c.__exit__(None, None, None)
                    logger.debug("exiting label_ctx %r", c)

            if s.fset_ctx is not None:
                s.fset_ctx.__exit__(None, None, None)
                logger.debug("exiting fset close %r", s.fset)

            if s.tab_ctx is not None:
                s.tab_ctx.__exit__(None, None, None)
                logger.debug("exiting tab close %r", s.tab)

            if s.tabview_ctx is not None:
                logger.debug("exiting tabview close")
                s.tabview_ctx.__exit__(None, None, None)
                parent.write(E.script(
                    '$(function(){ $( "#%s" ).tabs(); });' % s.tabview_id,
                    type="text/javascript"
                ))

        def complex_model_to_parent(self, ctx, cls, inst, parent, name,
                                                                      **kwargs):
            cls_attr = self.get_cls_attrs(cls)

            fieldset_attr = {'class': cls.get_type_name()}

            if cls_attr.fieldset is False or cls_attr.no_fieldset:
                return self._render_complex(ctx, cls, inst, parent, name,
                                                                False, **kwargs)

            if cls_attr.fieldset or cls_attr.no_fieldset is False:
                with parent.element('fieldset', fieldset_attr):
                    return self._render_complex(ctx, cls, inst, parent, name,
                                                                False, **kwargs)

            with parent.element('fieldset', fieldset_attr):
                return self._render_complex(ctx, cls, inst, parent, name,
                                                                 True, **kwargs)

        def _change_tab(self, ctx, parent, fti, s):
            global SOME_COUNTER

            if s.fset_ctx is not None:
                s.fset_ctx.__exit__(None, None, None)
                logger.debug("exiting fset tab %r ", s.prev_fset)

            s.fset_ctx = s.prev_fset = None

            if s.tab_ctx is not None:
                s.tab_ctx.__exit__(None, None, None)
                logger.debug("exiting tab %r", s.prev_tab)

            if s.prev_tab is None:
                logger.debug("entering tabview")
                s.tabview_id = 'tabview' + str(SOME_COUNTER)
                SOME_COUNTER += 1

                s.tabview_ctx = parent.element('div',
                                                    attrib={'id': s.tabview_id})
                s.tabview_ctx.__enter__()

                parent.write(self._gen_tab_headers(ctx, fti))

            logger.debug("entering tab %r", s.tab)

            attrib = {'id': s.tab.htmlid}
            attrib.update(s.tab.attrib)
            s.tab_ctx = parent.element('div', attrib)
            s.tab_ctx.__enter__()

            s.prev_tab = s.tab

        def _change_fset(self, ctx, parent, name, s):
            if s.fset_ctx is not None:
                s.fset_ctx.__exit__(None, None, None)
                logger.debug("exiting fset norm")

            logger.debug("entering fset %r", s.fset)
            s.fset_ctx = parent.element(s.fset.tag, s.fset.attrib)
            s.fset_ctx.__enter__()

            parent.write(E.legend(self.trd(s.fset.legend, ctx.locale, name)))
            s.prev_fset = s.fset

    return HtmlFormRoot


HtmlFormRoot = THtmlFormRoot(HtmlFormWidget)


class HtmlForm(HtmlFormRoot):
    def __init__(self, app=None, encoding='utf8',
                ignore_uncap=False, ignore_wrappers=False,
                cloth=None, cloth_parser=None, polymorphic=True, hier_delim='.',
                     doctype=None, label=True, placeholder=None,
                                         input_class=None, input_div_class=None,
                                     input_wrapper_class=None, label_class=None,
                                  action=None, method='POST', before_form=None):

        super(HtmlForm, self).__init__(app=app, encoding=encoding, doctype=doctype,
                     ignore_uncap=ignore_uncap, ignore_wrappers=ignore_wrappers,
                cloth=cloth, cloth_parser=cloth_parser, polymorphic=polymorphic,
                    hier_delim=hier_delim, label=label,
                    placeholder=placeholder, input_class=input_class,
                    input_div_class=input_div_class,
               input_wrapper_class=input_wrapper_class, label_class=label_class,
                          action=action, method=method, before_form=before_form)

        self.serialization_handlers = cdict({
            Date: self._check_simple(self.date_to_parent),
            Time: self._check_simple(self.time_to_parent),
            Uuid: self._check_simple(self.uuid_to_parent),
            File: self.file_to_parent,
            Fault: self.fault_to_parent,
            Array: self.array_type_to_parent,
            AnyUri: self._check_simple(self.any_uri_to_parent),
            AnyXml: self._check_simple(self.any_xml_to_parent),
            Integer: self._check_simple(self.integer_to_parent),
            Unicode: self._check_simple(self.unicode_to_parent),
            AnyHtml: self._check_simple(self.any_html_to_parent),
            Decimal: self._check_simple(self.decimal_to_parent),
            Boolean: self._check_simple(self.boolean_to_parent),
            Duration: self._check_simple(self.duration_to_parent),
            DateTime: self._check_simple(self.datetime_to_parent),
            ComplexModelBase: self.complex_model_to_parent,
        })

        self.use_global_null_handler = False

    def to_subprot(self, ctx, cls, inst, parent, name, subprot, **kwargs):
        if isinstance(subprot, HtmlFormWidget) or \
                                         not self.get_cls_attrs(cls).form_label:
            return subprot.subserialize(ctx, cls, inst, parent, name, **kwargs)

        return self._wrap_unknown_prot_with_label(ctx, cls, inst, parent,
                                                        name, subprot, **kwargs)

    @coroutine
    def _wrap_unknown_prot_with_label(self, ctx, cls, inst, parent, name,
                                                             subprot, **kwargs):
        attrib = self._gen_label_wrapper_class(ctx, cls, name)
        with parent.element('div', attrib):
            parent.write(E.label(self.trc(cls, ctx.locale, name)))
            ret = subprot.subserialize(ctx, cls, inst, parent, name, **kwargs)

        if isgenerator(ret):
            try:
                while True:
                    sv2 = (yield)
                    ret.send(sv2)
            except Break as b:
                try:
                    ret.throw(b)
                except StopIteration:
                    pass

    def unicode_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs, elt = self._gen_input_unicode(ctx, cls, inst, name, **kwargs)
        parent.write(self._wrap_with_label(ctx, cls, name, elt, **kwargs))

    def any_uri_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        if isinstance(inst, AnyUri.Value):
            # FIXME: this could have beeen handled *way* better.
            inst = inst.href
        self.unicode_to_parent(ctx, cls, inst, parent, name, **kwargs)

    @staticmethod
    @memoize_id
    def Tany_ml_to_parent(lxml_package):
        def _ml_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
            cls_attrs, elt = self._gen_input_textarea(ctx, cls, name, **kwargs)
            # TODO: Use something like ACE editor to generate a proper html editor

            if inst is not None:
                if cls_attrs.serialize_as in ('element', 'elt'):
                    if isinstance(inst, six.string_types):
                        inst = lxml_package.fromstring(inst)

                    elt.append(inst)

                else:
                    if not isinstance(inst, six.string_types):
                        inst = lxml_package.tostring(inst)

                    elt.text = inst

            parent.write(self._wrap_with_label(ctx, cls, name, elt, **kwargs))

        return _ml_to_parent

    any_html_to_parent = Tany_ml_to_parent.__func__(html)
    any_xml_to_parent = Tany_ml_to_parent.__func__(etree)

    def uuid_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        kwargs.update({'attr_override': {'max_len': 36}})
        cls_attrs, elt = self._gen_input_unicode(ctx, cls, inst, name, **kwargs)
        parent.write(self._wrap_with_label(ctx, cls, name, elt, **kwargs))

    def decimal_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)
        elt = self._gen_input(ctx, cls, inst, name, cls_attrs, **kwargs)
        elt.attrib['type'] = 'number'

        if D(cls.Attributes.fraction_digits).is_infinite():
            epsilon = self.HTML5_EPSILON
            elt.attrib['step'] = 'any'

        else:
            epsilon = 10 ** (-int(cls.Attributes.fraction_digits))
            elt.attrib['step'] = str(epsilon)

        self._apply_number_constraints(cls_attrs, elt, epsilon)

        parent.write(self._wrap_with_label(ctx, cls, name, elt, **kwargs))

    def boolean_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)
        elt = self._gen_input(ctx, cls, inst, name, cls_attrs,
                                           tag=self.HTML_CHECKBOX_TAG, **kwargs)

        if 'value' in elt.attrib:
            logger.debug("Removing value=%r from checkbox", elt.attrib['value'])
            del elt.attrib['value']
        elt.attrib['type'] = 'checkbox'

        if bool(inst):
            elt.attrib['checked'] = 'checked'

        wrap_label = HtmlFormWidget.WRAP_FORWARD \
                            if cls_attrs.label_position == 'left' \
                                                else HtmlFormWidget.WRAP_FORWARD

        elt = self._wrap_with_label(ctx, cls, name, elt,
                                                wrap_label=wrap_label, **kwargs)
        parent.write(elt)

    def date_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        ctx.protocol.assets.extend([('jquery',), ('jquery-ui', 'datepicker')])

        cls_attrs = self.get_cls_attrs(cls)
        elt = self._gen_input(ctx, cls, inst, name, cls_attrs, **kwargs)
        elt.attrib['type'] = 'text'

        if cls_attrs.format is None:
            data_format = 'yy-mm-dd'
        else:
            data_format = cls_attrs.format.replace('%Y', 'yy') \
                                          .replace('%m', 'mm') \
                                          .replace('%d', 'dd')

        code = [
            "$('#%(field_name)s').datepicker();",
            "$('#%(field_name)s').datepicker('option', 'dateFormat', '%(format)s');",
        ]

        if inst is None:
            script = self._format_js(code, field_name=elt.attrib['id'],
                                                             format=data_format)
        else:
            value = self.to_unicode(cls, inst)
            code.append(
                     "$('#%(field_name)s').datepicker('setDate', '%(value)s');")
            script = self._format_js(code, field_name=elt.attrib['id'],
                                                value=value, format=data_format)

        div = self._wrap_with_label(ctx, cls, name, elt, **kwargs)
        div.append(script)
        parent.write(div)

    def time_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        ctx.protocol.assets.extend([('jquery',), ('jquery-ui', 'datepicker'),
                                                        ('jquery-timepicker',)])

        cls_attrs = self.get_cls_attrs(cls)
        elt = self._gen_input(ctx, cls, inst, name, cls_attrs, **kwargs)
        elt.attrib['type'] = 'text'

        if cls_attrs.format is None:
            data_format = 'HH:MM:SS'
        else:
            data_format = cls_attrs.format.replace('%H', 'HH') \
                                          .replace('%M', 'MM') \
                                          .replace('%S', 'SS')

        code = [
            "$('#%(field_name)s').timepicker();",
            "$('#%(field_name)s').timepicker('option', 'timeFormat', '%(format)s');",
        ]

        if inst is None:
            script = self._format_js(code, field_name=elt.attrib['id'],
                                                             format=data_format)
        else:
            value = self.to_unicode(cls, inst)
            code.append(
                     "$('#%(field_name)s').timepicker('setDate', '%(value)s');")
            script = self._format_js(code, field_name=elt.attrib['id'],
                                                value=value, format=data_format)

        div = self._wrap_with_label(ctx, cls, name, elt, **kwargs)
        div.append(script)
        parent.write(div)

    @staticmethod
    def _split_datetime_format(f):
        # this is a pathetic attempt at splitting datetime format
        # to date and time formats. assumes date and time fields are not mixed

        lf = len(f)
        date_start_idx = min(
            _idxof(f, '%Y', lf), _idxof(f, '%m', lf), _idxof(f, '%d', lf)
        )
        date_end_idx = max(
            _idxof(f, '%Y', -1), _idxof(f, '%m', -1), _idxof(f, '%d', -1)
        ) + 2

        time_start_idx = min(
            _idxof(f, '%H', lf), _idxof(f, '%M', lf), _idxof(f, '%S', lf)
        )
        time_end_idx = max(
            _idxof(f, '%H', -1), _idxof(f, '%M', -1), _idxof(f, '%S', -1)
        ) + 2

        if date_end_idx <= time_start_idx:
            date_format = f[:date_end_idx].strip()
            time_format = f[time_start_idx:].strip()
        else:
            date_format = f[date_start_idx:].strip()
            time_format = f[:time_end_idx].strip()

        date_format = date_format.replace('%Y', 'yy') \
                                 .replace('%m', 'mm') \
                                 .replace('%d', 'dd') \

        time_format = time_format.replace('%H', 'HH') \
                                 .replace('%M', 'mm') \
                                 .replace('%S', 'ss')

        return date_format, time_format

    def datetime_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        # see: http://trentrichardson.com/examples/timepicker/
        ctx.protocol.assets.extend([('jquery',), ('jquery-ui', 'datepicker'),
                                                        ('jquery-timepicker',)])

        cls_attrs = self.get_cls_attrs(cls)
        elt = self._gen_input(ctx, cls, None, name, cls_attrs, **kwargs)
        elt.attrib['type'] = 'text'

        dt_format = self._get_datetime_format(cls_attrs)
        if dt_format is None:
            date_format, time_format = 'yy-mm-dd', 'HH:mm:ss'
        else:
            date_format, time_format = \
                                self._split_datetime_format(cls_attrs.dt_format)

        code = [
            "var t=$('#%(field_name)s');",
            "t.datetimepicker({",
            "    dateFormat: '%(date_format)s',",
            "    timeFormat: '%(time_format)s'",
            "});",
        ]

        if inst is None:
            script = self._format_js(code, field_name=elt.attrib['id'],
                               date_format=date_format, time_format=time_format)
        else:
            value = self.to_unicode(cls, inst)

            code.append(
                "t.datetimepicker('setTime', '%(value)s');")

            script = self._format_js(code, field_name=elt.attrib['id'],
                  date_format=date_format, time_format=time_format, value=value)

        div = self._wrap_with_label(ctx, cls, name, elt, **kwargs)
        div.append(script)
        parent.write(div)

    def integer_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)
        elt = self._gen_input(ctx, cls, inst, name, cls_attrs, **kwargs)

        if elt.tag == self.HTML_INPUT:
            # no point in applying constraints for <select> tags
            elt.attrib['type'] = 'number'
            self._apply_number_constraints(cls_attrs, elt, epsilon=1)

        parent.write(self._wrap_with_label(ctx, cls, name, elt, **kwargs))

    def file_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)

        elt = self._gen_input(ctx, cls, '', name, cls_attrs, **kwargs)
        elt.attrib['type'] = 'file'
        elt.attrib['style'] = 'display:inline-block'

        wrapped_elt = self._wrap_with_label(ctx, cls, name, elt, **kwargs)

        if inst is None:
            inst = ''

        elif isinstance(inst, (File.Value, FileData)):
            inst = '<data type="%s" size="%d">' % \
                                   (inst.type, sum([len(d) for d in inst.data]))

        else:
            inst = '<data size="%d">' % sum([len(d) for d in inst])

        if len(inst) > 0:
            parelt = elt.getparent()
            parelt.append(E.span(inst, " | ", elt,
                                               style="font-family: monospace;"))

        parent.write(wrapped_elt)

    # TODO: finish this
    def duration_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attrs = self.get_cls_attrs(cls)
        elt = self._gen_input(ctx, cls, inst, name, cls_attrs, **kwargs)

        parent.write(self._wrap_with_label(ctx, cls, name, elt, **kwargs))

    def array_type_to_parent(self, ctx, cls, inst, parent, name=None, **kwargs):
        v = next(iter(cls._type_info.values()))
        return self.to_parent(ctx, v, inst, parent, name, **kwargs)

    def fault_to_parent(self, ctx, cls, inst, parent, name, **kwargs):
        cls_attr = self.get_cls_attrs(cls)

        with parent.element('div', {"class": cls.get_type_name()}):
            parent.write(
                E.div(inst.faultcode, **{"class": "faultcode"}),
                E.div(inst.faultstring, **{"class": "faultstring"}),
                E.div(inst.faultactor, **{"class": "faultactor"}),
            )

            if isinstance(inst.detail, etree._Element):
                parent.write(E.pre(etree.tostring(inst.detail, pretty_print=True)))

            # add other nonstandard fault subelements with get_members_etree
            self._write_members(ctx, cls, inst, parent)
            # no need to track the returned generator because we expect no
            # PushBase instance here.

    @coroutine
    def _push_to_parent(self, ctx, cls, inst, parent, name, parent_inst=None,
                        label_attrs=None, parent_key=None, no_label=False,
                        **kwargs):
        i = -1
        key = self.selsafe(name)
        attr = self.get_cls_attrs(cls)
        while True:
            with parent.element('div', {"class": key}):
                sv = (yield)

                i += 0
                new_key = '%s[%09d]' % (key, i)
                ret = self.to_parent(ctx, cls, sv, parent, new_key,
                                                from_arr=True, **kwargs)

                if isgenerator(ret):
                    try:
                        while True:
                            sv2 = (yield)
                            ret.send(sv2)
                    except Break as e:
                        try:
                            ret.throw(e)
                        except StopIteration:
                            pass

                        if not attr.no_write:
                            parent.write(E.button('+', **{
                                "class": key + "_btn_add",
                                'type': 'button'
                            }))
                            parent.write(E.button('-', **{
                                "class": key + "_btn_del",
                                'type': 'button'
                            }))

        if not attr.no_write:
            _gen_array_js(parent, key)

    @coroutine
    def _pull_to_parent(self, ctx, cls, inst, parent, name, parent_inst=None,
                        label_attrs=None, parent_key=None, no_label=False,
                        **kwargs):
        key = self.selsafe(name)
        attr = self.get_cls_attrs(cls)

        if inst is None:
            inst = []

        for i, subval in enumerate(inst):
            new_key = '%s[%09d]' % (key, i)
            with parent.element('div', {"class": key}):
                kwargs['from_arr'] = True
                ret = self.to_parent(ctx, cls, subval, parent, new_key,
                                         parent_inst=parent_inst, no_label=True,
                                         **kwargs)
                if not attr.no_write:
                    parent.write(E.button('+', **{
                                  "class": key + "_btn_add", 'type': 'button'}))
                    parent.write(E.button('-', **{
                                  "class": key + "_btn_del", 'type': 'button'}))

                if isgenerator(ret):
                    try:
                        while True:
                            sv2 = (yield)
                            ret.send(sv2)
                    except Break as b:
                        try:
                            ret.throw(b)
                        except StopIteration:
                            pass

        if not attr.no_write:
            _gen_array_js(parent, key)

    def array_to_parent(self, ctx, cls, inst, parent, name, parent_inst=None,
                        label_attrs=None, parent_key=None, no_label=False,
                        **kwargs):

        key = self.selsafe(name)

        with parent.element('div', {"id": key + "_container", 'class':'array'}):
            if isinstance(inst, PushBase):
                return self._push_to_parent(ctx, cls, inst, parent, name,
                                     parent_inst=parent_inst, label_attrs=None,
                                     parent_key=None, no_label=False, **kwargs)

            else:
                return self._pull_to_parent(ctx, cls, inst, parent, name,
                                     parent_inst=parent_inst, label_attrs=None,
                                     parent_key=None, no_label=False, **kwargs)



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from neurons.base.screen import ScreenBase

from spyne import Unicode, ComplexModel, XmlAttribute, Array, AnyUri, Boolean, \
    Integer, XmlData, Date, AnyHtml, DateTime, Time


class HtmlImport(ComplexModel):
    href = XmlAttribute(Unicode)


class IronAjax(ComplexModel):
    url = XmlAttribute(AnyUri)


class HtmlElementBase(ComplexModel):
    id = XmlAttribute(Unicode)
    class_ = XmlAttribute(Unicode(sub_name='class'))


class HtmlFormElementBase(HtmlElementBase):
    name = XmlAttribute(Unicode)
    disabled = XmlAttribute(Boolean)
    readonly = XmlAttribute(Boolean)


class PaperItem(HtmlElementBase):
    class Attributes(ComplexModel.Attributes):
        sub_name = 'paper-item'

    label = XmlData(Unicode)
    value = XmlAttribute(Unicode)


class PaperListbox(HtmlElementBase):
    class Attributes(ComplexModel.Attributes):
        sub_name = 'paper-listbox'

    items = Array(PaperItem, wrapped=False)
    selected = XmlAttribute(Integer)


class PaperDropdownMenu(HtmlFormElementBase):
    class Attributes(ComplexModel.Attributes):
        sub_name = 'paper-dropdown-menu'

    listbox = PaperListbox
    noink = XmlAttribute(Boolean)
    label = XmlAttribute(Unicode)
    no_animations = XmlAttribute(Boolean(sub_name='no-animations'))
    always_float_label = XmlAttribute(Boolean(sub_name='always-float-label'))


class NeuronsDateTimePicker(HtmlElementBase):
    class Attributes(ComplexModel.Attributes):
        sub_name = 'neurons-datetime-picker'

    min = DateTime
    max = DateTime
    no_date = XmlAttribute(Boolean(sub_name='no-date'))
    no_time = XmlAttribute(Boolean(sub_name='no-time'))
    format = XmlAttribute(Unicode)

    name = XmlAttribute(Unicode)
    label = XmlAttribute(Unicode)
    required = XmlAttribute(Boolean)
    readonly = XmlAttribute(Boolean)
    # always_float_label = XmlAttribute(Boolean(sub_name='always-float-label'))


class Template(ComplexModel):
    data = XmlData(AnyHtml)


class IronDataTableColumn(ComplexModel):
    class Attributes(ComplexModel.Attributes):
        sub_name = 'data-table-column'

    name = XmlAttribute(Unicode)
    template = Template


class IronDataTable(ComplexModel):
    class Attributes(ComplexModel.Attributes):
        sub_name = 'iron-data-table'

    items = XmlAttribute(Unicode)
    columns = Array(IronDataTableColumn, wrapped=False)


class NeuronsArray(HtmlFormElementBase):
    class Attributes(ComplexModel.Attributes):
        sub_name = 'neurons-array'

    label = XmlAttribute(Unicode)
    columns = Array(IronDataTableColumn, wrapped=False)
    arg_map = XmlAttribute(Unicode(sub_name='arg-map'))


class NeuronsComplexHref(HtmlFormElementBase):
    class Attributes(ComplexModel.Attributes):
        sub_name = 'neurons-complex-href'

    label = XmlAttribute(Unicode)
    base_href = XmlAttribute(Unicode(sub_name='base-href'))
    param_whitelist = XmlAttribute(Unicode(sub_name='param-whitelist'))
    attr_item_value = XmlAttribute(Unicode(sub_name='attr-item-value'))
    attr_item_label = XmlAttribute(Unicode(sub_name='attr-item-label'))
    need_parent_params = XmlAttribute(Boolean(sub_name='need-parent-params'))
    value = XmlAttribute(Boolean(sub_name='value'))


class NeuronsComplexDropdown(HtmlFormElementBase):
    class Attributes(ComplexModel.Attributes):
        sub_name = 'neurons-complex-dropdown'

    label = XmlAttribute(Unicode)
    readonly = XmlAttribute(Boolean)
    always_float_label = XmlAttribute(Boolean)
    data_source = XmlAttribute(Unicode(sub_name='data-source'))
    param_whitelist = XmlAttribute(Unicode(sub_name='param-whitelist'))
    attr_item_value = XmlAttribute(Unicode(sub_name='attr-item-value'))
    attr_item_label = XmlAttribute(Unicode(sub_name='attr-item-label'))
    need_parent_params = XmlAttribute(Boolean(sub_name='need-parent-params'))
    value = XmlAttribute(Boolean(sub_name='value'))


class PaperCheckbox(HtmlFormElementBase):
    class Attributes(ComplexModel.Attributes):
        sub_name = 'paper-checkbox'

    label = Unicode


class PaperInput(HtmlFormElementBase):
    class Attributes(ComplexModel.Attributes):
        sub_name = 'paper-input'

    type = XmlAttribute(Unicode)
    min = XmlAttribute(Integer)
    max = XmlAttribute(Integer)
    minlength = XmlAttribute(Integer)
    maxlength = XmlAttribute(Integer)
    placeholder = XmlAttribute(Unicode)
    label = XmlAttribute(Unicode)
    always_float_label = XmlAttribute(Boolean(sub_name='always-float-label'))
    required = XmlAttribute(Boolean)
    auto_validate = XmlAttribute(Boolean(sub_name='auto-validate'))
    error_message = XmlAttribute(Unicode(sub_name='error-message'))
    pattern = XmlAttribute(Unicode)


class PolymerComponent(ComplexModel):
    style = Unicode
    getter = IronAjax
    putter = IronAjax
    definition = Unicode
    dependencies = Array(HtmlImport, wrapped=False)
    dom_module_id = Unicode(sub_name="id")


XML_TO_JSOBJ = """
window.neurons.xml_to_jsobj = function (children, array_tags) {
    function is_array(o) {
        return Object.prototype.toString.apply(o) === '[object Array]';
    }

    function parse_node(xml_node, result, parnames) {
        if (xml_node.nodeName == "#text") {
            var str = xml_node.nodeValue.trim();
            if (str != "") {
                result[parnames.join('.')] = str;
            }

            return;
        }

        var json_node = {};
        var curname = xml_node.nodeName.toLowerCase();

        if (xml_node.attributes) {
            var length = xml_node.attributes.length;
            for (var i = 0; i < length; i++) {
                var attribute = xml_node.attributes[i];
                var attrname = attribute.nodeName.toLowerCase();
                var names = parnames.concat([attrname]);
                json_node[names.join('.')] = attribute.nodeValue;
            }
        }

        var length = xml_node.childNodes.length;
        for (var i = 0; i < length; i++) {
            parse_node(xml_node.childNodes[i], result,
                                                    parnames.concat([curname]));
        }
    }

    var result = {};
    for (var i = 0, l = children.length; i < l; ++i) {
        parse_node(children[i], result, []);
    }

    return result;
}
"""


class PolymerScreen(ScreenBase):
    JQUERY_URL = "/static/jquery-1.12.4.min.js"

    def __init__(self, ctx, *args, **kwargs):
        super(PolymerScreen, self).__init__(ctx, *args, **kwargs)
        self._have_xml_to_jsobj = False

    def with_xml_to_jsobj(self):
        if not self._have_xml_to_jsobj:
            self.with_namespace()

            self.append_script(XML_TO_JSOBJ)

            self._have_xml_to_jsobj = True

        return self



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from __future__ import print_function, absolute_import, unicode_literals

import logging
logger = logging.getLogger(__name__)

import os, sys
import getpass
import traceback

from os import access
from uuid import uuid1
from warnings import warn
from os.path import isfile, abspath, dirname, getsize
from argparse import Action
from pwd import getpwnam, getpwuid
from grp import getgrnam

from spyne import ComplexModel, Boolean, ByteArray, Uuid, Unicode, Array, \
    String, UnsignedInteger16, M, Integer32, ComplexModelMeta, ComplexModelBase
from spyne.protocol import ProtocolBase
from spyne.protocol.yaml import YamlDocument
from spyne.util import six
from spyne.util.dictdoc import get_object_as_yaml, get_yaml_as_object

from neurons import is_reactor_thread, CONFIG_FILE_VERSION
from neurons.daemon.cli import spyne_to_argparse
from neurons.daemon.daemonize import daemonize_do

from neurons.daemon.config import LOGLEVEL_STR_MAP
from neurons.daemon.config._wdict import wdict, Twrdict
from neurons.daemon.config.limits import LimitsChoice

from neurons.daemon.config import FILE_VERSION_KEY
from neurons.daemon.config.endpoint import Service, Server
from neurons.daemon.config.logutils import Logger, Trecord_as_string, \
    TDynamicallyRotatedLog, TTwistedHandler
from neurons.daemon.config.store import RelationalStore, StorageInfo


_some_prot = ProtocolBase()

meminfo = None

def update_meminfo():
    """Call this when the process pid changes."""

    global meminfo

    try:
        import psutil
        process = psutil.Process(os.getpid())
        try:  # psutil 2
            meminfo = process.get_memory_info
        except AttributeError:  # psutil 3
            meminfo = process.memory_info

        del process

    except ImportError:
        pass


update_meminfo()


class _SetStaticPathAction(Action):
    def __init__(self, option_strings, dest, const=None, help=None):
        super(_SetStaticPathAction, self).__init__(nargs=1, const=const,
                            option_strings=option_strings, dest=dest, help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        self.const.path = abspath(values[0])


class AlertDestination(ComplexModel):
    type = Unicode(default='email', values=['email'])


EmailAddress = Unicode(pattern=r"[^@\s]+@[^@\s]+")


class EmailAlert(AlertDestination):
    host = Unicode(default='localhost')
    port = UnsignedInteger16(default=25)
    user = Unicode
    sender = Unicode(default='Joe Developer <somepoorhuman@spyne.io>')
    envelope_from = EmailAddress
    password = Unicode
    recipients = M(Array(M(EmailAddress)))


class ConfigBaseMeta(ComplexModelMeta):
    def __init__(self, *args, **kwargs):
        super(ConfigBaseMeta, self).__init__(*args, **kwargs)

        for k, v in list(self._type_info.items()):
            if getattr(v.Attributes, 'no_config', None):
                self._type_info[k] = \
                                  v.customize(pa={YamlDocument: dict(exc=True)})


@six.add_metaclass(ConfigBaseMeta)
class ConfigBase(ComplexModelBase):
    pass


class Daemon(ConfigBase):
    """This is a custom daemon with only pid files, forking, logging.
    No setuid/setgid support is implemented.
    """

    LOGGING_DEBUG_FORMAT = \
                     "%(l)s %(r)s | %(module)12s:%(lineno)-4d | %(message)s"
    LOGGING_DEVEL_FORMAT = "%(l)s | %(module)12s:%(lineno)-4d | %(message)s"
    LOGGING_PROD_FORMAT = \
               "%(l)s %(asctime)s | %(module)12s:%(lineno)-4d | %(message)s"

    _type_info = [
        ('name', Unicode(no_cli=True, help="Daemon Name")),
        ('help', Boolean(no_config=True, short='h',
                                       help="Show this help message and exit")),
        ('file_version', Integer32(no_cli=True, sub_name=FILE_VERSION_KEY,
                                                  default=CONFIG_FILE_VERSION)),
        ('uuid', Uuid(
            no_cli=True,
            help="Daemon uuid. Regenerated every time a new config file is "
                 "written. It could come in handy.")),

        ('secret', ByteArray(
            no_cli=True,
            help="Secret key for signing cookies and other stuff.")),

        ('daemonize', Boolean(
            default=False,
            help="Daemonizes before everything else.")),

        ('workdir', Unicode(
            help="The daemon workdir. The daemon won't boot if this doesn't "
                 "exist and can't be created.")),
        ('uid', Unicode(
            help="The daemon user. You need to start the server as a "
                 "privileged user for this to work.")),
        ('gid', Unicode(
            help="The daemon group. You need to start the server as a "
                 "privileged user for this to work.")),

        ('gids', Array(Unicode,
            help="Additional groups for the daemon. Use an empty list to drop"
                 "all additional groups.")),

        ('limits', LimitsChoice.customize(not_wrapped=True,
            help="Process limits.")),

        ('pid_file', String(
            help="The path to a text file that contains the pid of the "
                 "daemonized process.")),

        ('logger_dest', String(
            help="The path to the log file. The server won't daemonize "
                 "without this. Converted to an absolute path if not.")),

        ('logger_dest_rotation_period', Unicode(
            values=['DAILY', 'WEEKLY', 'MONTHLY'],
            help="Logs rotation period")),

        ('logger_dest_rotation_compression', Unicode(
            values=['gzip'],
            help="Logs rotation compression")),

        ('version', Boolean(help="Show version", no_file=True)),

        ('bootstrap', Boolean(
            no_file=True,
            help="Bootstrap the application instead of starting it. "
                 "Create database schema, insert initial data, etc.")),

        ('log_rss', Boolean(default=False,
            help="Prepend resident set size to all logging messages. "
                 "Requires psutil")),
        ('log_protocol', Boolean(default=False,
                                              help="Log protocol operations.")),
        ('log_model', Boolean(default=False, help="Log model operations.")),
        ('log_cloth', Boolean(default=False, help="Log cloth generation.")),
        ('log_interface', Boolean(default=False,
                                          help="Log interface build process.")),

        ('write_config', Boolean(no_file=True,
                                   help="Write configuration file and exit.")),

        ('alert_dests', Array(AlertDestination, default=[])),

        ('shell', Boolean(
            no_file=True, default=False,
            help="Drop to IPython shell. Useful for trying ORM stuff")),
        ('ikernel', Boolean(
            no_file=True, default=False,
            help="Start IPython kernel.")),

        ('autoreload', Boolean(
            default=False,
            help="Auto-relaunch daemon process when one of "
                 "the source files change.")),

        ('dry_run', Boolean(
            no_file=True, default=False,
            help="Do everything up until the reactor start and "
                 "exit instead of starting the reactor.")),

        ('debug', Boolean(default=False)),
        ('debug_reactor', Boolean(default=False)),

        ('_services', Array(Service, sub_name='services')),
        ('_loggers', Array(Logger, sub_name='loggers')),
    ]

    # FIXME: We need all this hacky magic with custom constructor and properties
    #        because Spyne doesn't support custom containers
    def __init__(self, *args, **kwargs):
        super(Daemon, self).__init__(*args, **kwargs)
        self._boot_messaged = False

        services = kwargs.get('services', None)
        if services is not None:
            self.services = services
        if self.services is None:
            self.services = Twrdict(self, 'name')()
        self._set_parent_of_children(self.services)

        loggers = kwargs.get('loggers', None)
        if loggers is not None:
            self.loggers = loggers
        if self.loggers is None:
            self.loggers = wdict()
        self._set_parent_of_children(self.loggers)

    def _set_parent_of_children(self, wrd):
        for v in wrd.values():
            v.set_parent(self)

    def _parse_overrides(self):
        for s in self.services.values():
            s._parse_overrides()

    @property
    def _services(self):
        if self.services is not None:
            for k, v in self.services.items():
                if v is not None:
                    v.name = k

            return self.services.values()

        self.services = Twrdict(self, 'name')()
        return []

    @_services.setter
    def _services(self, what):
        self.services = what
        if what is not None:
            self.services = Twrdict(self, 'name')([(s.name, s) for s in what])

    @property
    def _loggers(self):
        if self.loggers is not None:
            for k, v in self.loggers.items():
                v.name = k

            return self.loggers.values()

        self.loggers = wdict()
        return []

    @_loggers.setter
    def _loggers(self, what):
        self.loggers = what
        if what is not None:
            self.loggers = wdict([(s.path, s) for s in what])

    @classmethod
    def gen_secret(cls):
        return [os.urandom(64)]

    @classmethod
    def gen_uuid(cls):
        return uuid1()

    @classmethod
    def get_default(cls, daemon_name):
        return cls(
            debug=True,
            uuid=cls.gen_uuid(),
            secret=cls.gen_secret(),
            name=daemon_name,
            log_rss=True,
            workdir=os.getcwd(),
            _loggers=[
                Logger(path='.', level='DEBUG', format=cls.LOGGING_DEVEL_FORMAT),
            ],
        )

    def _clear_other_observers(self, publisher, observer):
        from twisted.logger import LimitedHistoryLogObserver, LogPublisher

        # FIXME: Remove Limited History Observer in a supported way.
        logger.debug("Looking for rogue observers in %r", publisher._observers)

        for o in publisher._observers:
            if isinstance(o, LogPublisher):
                self._clear_other_observers(o, observer)

            elif isinstance(o, LimitedHistoryLogObserver):
                publisher.removeObserver(o)
                o.replayTo(observer)
                logger.debug("Removing observer %r", o)

            else:
                logger.debug("Leaving alone observer %r", o)

    def sanitize(self):
        if self.logger_dest is not None:
            self.logger_dest = abspath(self.logger_dest)
        if self.pid_file is not None:
            self.pid_file = abspath(self.pid_file)

    def apply_logging(self):
        # We're using twisted logging only for IO.
        from twisted.logger import FileLogObserver
        from twisted.logger import globalLogPublisher

        loggers = {}

        if self.logger_dest is not None:
            DynamicallyRotatedLog = TDynamicallyRotatedLog(self,
                                          self.logger_dest_rotation_compression)

            self.logger_dest = abspath(self.logger_dest)
            if access(dirname(self.logger_dest), os.R_OK | os.W_OK | os.X_OK):
                log_dest = DynamicallyRotatedLog.fromFullPath(self.logger_dest)

            else:
                warn("'%s' is not accessible. We need at least rw- on "
                               "it to rotate logs." % dirname(self.logger_dest))

                log_dest = open(self.logger_dest, 'w+')

            if self.debug:
                formatter = logging.Formatter(self.LOGGING_DEBUG_FORMAT)
            else:
                formatter = logging.Formatter(self.LOGGING_PROD_FORMAT)

            os.chown(self.logger_dest, self.get_uid(), self.get_gid())

        else:
            if self.debug:
                formatter = logging.Formatter(self.LOGGING_DEBUG_FORMAT)
            else:
                formatter = logging.Formatter(self.LOGGING_DEVEL_FORMAT)
            log_dest = sys.stdout

            try:
                callers = {
                    func_name for file_name, line_number, func_name, code in
                                                      traceback.extract_stack()}

                if 'pytest_cmdline_main' in callers:
                    print("colorama not loaded because pytest was detected.")

                else:
                    import colorama
                    colorama.init()
                    if self.debug:
                        print("colorama loaded.")

            except Exception as e:
                if self.debug:
                    print("coloarama not loaded: %r" % e)

        record_as_string = Trecord_as_string(formatter)
        observer = FileLogObserver(log_dest, record_as_string)
        globalLogPublisher.addObserver(observer)
        self._clear_other_observers(globalLogPublisher, observer)

        TwistedHandler = TTwistedHandler(self, loggers, meminfo)

        handler = TwistedHandler()
        handler.setFormatter(formatter)
        logging.getLogger().addHandler(handler)

        self.pre_logging_apply()

        for l in self.loggers.values() or []:
            l.set_parent(self)
            l.apply()

        self.boot_message()

        return self

    def boot_message(self):
        if self._boot_messaged:
            return
        self._boot_messaged = True

        import spyne
        import neurons
        import twisted
        import sqlalchemy

        logger.info("Booting daemon '%s' with spyne-%s, neurons-%s, "
            "sqlalchemy-%s and twisted-%s.", self.name,
                                spyne.__version__, neurons.__version__,
                                sqlalchemy.__version__, twisted.version.short())

    @staticmethod
    def hello_darkness_my_old_friend():
        logger.info("If you see this, it means something else has also "
                "initialized the logging subsystem. This means you will get "
                                                     "duplicate logging lines.")

    def pre_logging_apply(self):
        if self.debug:
            print("Root logger level = %s" %
                                    LOGLEVEL_STR_MAP[logging.getLogger().level])

        self.hello_darkness_my_old_friend()

    def pre_limits_apply(self):
        """Override this function in case a resource-intensive task needs to be
        done prior to daemon boot. This is generally a bad idea but hey, who are
        we to tell you what to do?.. :)"""

    def apply_limits(self):
        if self.limits is not None:
            self.pre_limits_apply()
            self.limits.apply()

    def apply_listeners(self):
        dl = []

        for s in self._services:
            if isinstance(s, Server):
                if not s.disabled:
                    dl.append(s.listen())

        return dl

    def get_gid(self):
        if self.gid is None:
            return -1

        gid = self.gid
        if not isinstance(gid, int):
            return getgrnam(self.gid).gr_gid

        return gid

    def get_uid(self):
        if self.uid is None:
            return -1

        uid = self.uid
        if not isinstance(uid, int):
            return getpwnam(self.uid).pw_uid

        return uid

    def apply_uidgid(self):
        if self.gid is not None:
            gid = self.gid
            if not isinstance(gid, int):
                gid = getgrnam(self.gid).gr_gid

            os.setgid(gid)
            os.setegid(gid)
            os.setgroups([])

        if self.gids is not None:
            os.setgroups([])
            os.setgroups([gid if isinstance(gid, int)
                                else getgrnam(gid).gr_gid for gid in self.gids])

        if self.uid is not None:
            uid = self.uid
            if isinstance(uid, int):
                pw = getpwuid(uid)
            else:
                pw = getpwnam(uid)

            if self.gid is None:
                os.setgid(pw.pw_gid)
                if self.gids is None:
                    os.setgroups([])

            os.setuid(pw.pw_uid)
            os.seteuid(pw.pw_uid)

    def apply_daemonize(self):
        # Daemonization won't work if twisted is imported before fork().
        # It's best to know this in advance or you'll have to deal with daemons
        # that work perfectly well in development environments but won't boot
        # in production ones, solely because daemonization involves closing of
        # previously open file descriptors.
        if ('twisted' in sys.modules):
            import twisted
            raise Exception(
                 "Twisted is already imported from {}".format(twisted.__file__))

        self.sanitize()
        if self.daemonize:
            assert self.logger_dest, "Refusing to start without any log " \
                            "output. Please set logger_dest in the config file."

            workdir = self.workdir
            if workdir is None:
                workdir = '/'
            daemonize_do(workdir=workdir)

            update_meminfo()

        else:
            if self.workdir is not None:
                os.chdir(self.workdir)
                logger.debug("Change working directory to: %s", self.workdir)


    def apply_pid_file(self):
        if self.pid_file is not None:
            pid = os.getpid()
            with open(self.pid_file, 'w') as f:
                f.write(str(pid))
                logger.debug("Pid file is at: %r", self.pid_file)

    def apply(self, daemonize=True):
        """Daemonizes the process if requested, then sets up logging and pid
        files.
        """

        # FIXME: this should really be "may_daemonize"
        if daemonize:
            self.apply_daemonize()
            self.apply_pid_file()

        self.apply_logging()

        if self.debug:
            import twisted.internet.base
            twisted.internet.base.DelayedCall.debug = True

            import twisted.internet.defer
            twisted.internet.defer.setDebugging(True)

            logger.info("Enabled debugging for "
                 "twisted.internet.base.DelayedCall and twisted.internet.defer")

        if daemonize:
            self.apply_limits()
            self.apply_listeners()
            self.apply_uidgid()

        return self

    def add_reactor_checks(self):
        """Logs warnings when stuff that could be better off in a dedicated
        thread is found running inside the reactor thread."""

        from sqlalchemy import event

        def before_cursor_execute(conn, cursor, statement, parameters, context,
                                                                   executemany):
            if is_reactor_thread():
                logger.warning("SQL query found inside reactor thread. "
                                    "Statement: %s Traceback: %s",
                                   statement, ''.join(traceback.format_stack()))

        for store in self._stores:
            if not isinstance(store, RelationalStore):
                continue

            engine = store.itself.engine
            event.listen(engine, "before_cursor_execute", before_cursor_execute)

            logger.info("Installed reactor warning hook for engine %s.", engine)

    @classmethod
    def _apply_custom_attributes(cls):
        fti = cls.get_flat_type_info(cls)
        for k, v in sorted(fti.items(), key=lambda i: i[0]):
            attrs = _some_prot.get_cls_attrs(v)
            if attrs.no_file == True:
                v.Attributes.prot_attrs = {YamlDocument: dict(exc=True)}

    @classmethod
    def parse_config(cls, daemon_name, argv=None):
        cls._apply_custom_attributes()
        file_name = abspath('%s.yaml' % daemon_name)

        argv_parser = spyne_to_argparse(cls, ignore_defaults=True)
        cli = {}
        if argv is not None and len(argv) > 1:
            cli = dict(argv_parser.parse_args(argv[1:]).__dict__.items())
            if cli['config_file'] is not None:
                file_name = abspath(cli['config_file'])
                del cli['config_file']

        exists = isfile(file_name) and os.access(file_name, os.R_OK)
        if exists and getsize(file_name) > 0:
            s = open(file_name, 'rb').read()

        else:
            if not access(dirname(file_name), os.R_OK | os.W_OK):
                raise Exception("File %r can't be created in %r" %
                                                (file_name, dirname(file_name)))
            s = b""

        retval = cls.parse_config_string(s, daemon_name, cli)
        retval.config_file = file_name
        return retval

    @classmethod
    def parse_config_string(cls, s, daemon_name, cli=None):
        if cli is None:
            cli = {}

        retval = cls.get_default(daemon_name)
        if len(s) > 0:
            s = retval._migrate_impl(s)
            try:
                retval = get_yaml_as_object(s, cls,
                                             validator='soft', polymorphic=True)
            except Exception as e:
                logger.error("Error parsing %s", repr(e))
                logger.exception(e)
                logger.error("File: %s", s)
                raise

        retval._parse_overrides()

        for k, v in cli.items():
            if not v in (None, False):
                setattr(retval, k, v)

        import neurons.daemon
        neurons.daemon.config_data = retval

        return retval

    def do_write_config(self):
        open(self.config_file, 'wb').write(get_object_as_yaml(self,
                                              self.__class__, polymorphic=True))

    def get_email_alert_addresses(self):
        retval = []

        if self.alert_dests is not None:
            for dest in self.alert_dests:
                if isinstance(dest, EmailAlert):
                    retval.extend(dest.recipients)

        return retval

    def _migrate_impl(self, s):
        """Migrates old config files to new ones. Takes old config file as
        string and returns the new one.

        This is not supposed to be overridden. Override migrate_dict if you
        want add migration
        """

        import yaml
        config_dict = yaml.load(s)
        key, = config_dict.keys()

        config_root = config_dict[key]
        config_version = config_root.get(FILE_VERSION_KEY, None)

        if config_version is None:
            # Perform relational store migration
            try:
                z = config_root['stores'][0]['Relational']

            except KeyError:
                pass

            else:
                del config_root['stores'][0]['Relational']
                config_root['stores'][0]['RelationalStore'] = z

            # use the minimum version number to have all the migration
            # operations be run
            config_version = 1

        else:
            self.migrate(config_dict, config_version)

        if config_version < 2:
            print("Performing config file migration "
                                  "from version %d to 2..." % (config_version,))

            for k1, v1 in config_root.items():
                if k1 == 'services':
                    for v2 in v1:
                        for k3, v3 in v2.items():
                            old = new = k3

                            if k3 == 'Listener':
                                new = 'Server'

                            if k3 == 'SslListener':
                                new = 'SslServer'

                            if k3 == 'IpcListener':
                                new = 'IpcServer'

                            if k3 == 'HttpListener':
                                new = 'HttpServer'

                            if k3 == 'DowserListener':
                                new = 'DowserServer'

                            if k3 == 'WsgiListener':
                                new = 'WsgiServer'

                            if k3 == 'StaticFileListener':
                                new = 'StaticFileServer'

                            if old != new:
                                v2[new] = v2[old]
                                print("\tRenamed service type %s to %s" %
                                                                     (old, new))
                                del v2[old]

            print("Config file migration from version %d to 2 "
                                              "successful." % (config_version,))

        config_root[FILE_VERSION_KEY] = CONFIG_FILE_VERSION
        return yaml.dump(config_dict, indent=4, default_flow_style=False)

    def migrate(self, config_dict, config_version):
        """Makes in-place changes in the config_dict before it's parsed and
        applied by the configuration engine."""

        # example
        if config_version < (0, 6, 0):
            key, = config_dict.keys()
            _ = config_dict[key]
            # do something with the dict

        pass


class ServiceDaemon(Daemon):
    """This is a daemon with data stores."""
    DEFAULT_DB_NAME = None

    _type_info = [
        ('write_wsdl', Unicode(
            help="Write Wsdl document(s) to the given directory and exit. "
                 "It is created if missing", no_file=True, metavar='WSDL_DIR')),

        ('write_xsd', Unicode(
            help="Write Xml Schema document(s) to given directory and exit. "
                  "It is created if missing", no_file=True, metavar='XSD_DIR')),

        ('log_orm', Boolean(default=False,
                                        help="Log SQLAlchemy ORM operations.")),
        ('log_queries', Boolean(default=False, help="Log SQL queries.")),
        ('log_results', Boolean(default=False,
                         help="Log SQL query results in addition to queries.")),
        ('log_dbconn', Boolean(default=False,
            help="Prepend number of open database connections to all "
                                                          "logging messages.")),
        ('log_sqlalchemy', Boolean(default=False,
                                help="Log SQLAlchemy messages in debug level")),

        ('drop_all_tables', Boolean(help="Drops all tables in the database.",
                                                                 no_file=True)),

        ('main_store', Unicode(help="The name of the store for binding "
                                          "neurons.TableModel's metadata to.")),

        ('gen_data', Boolean(help="Generates random data", no_file=True)),
        ('_stores', Array(StorageInfo, sub_name='stores')),
    ]

    # FIXME: We need all this hacky magic with custom constructor and properties
    # because Spyne doesn't support custom containers
    def __init__(self, *args, **kwargs):
        super(ServiceDaemon, self).__init__(*args, **kwargs)

        stores = kwargs.get('stores', None)
        if stores is not None:
            self.stores = stores
        if not hasattr(self, 'stores') or self.stores is None:
            self.stores = wdict()
        self._set_parent_of_children(self.stores)

    @property
    def _stores(self):
        if self.stores is not None:
            for k, v in self.stores.items():
                v.name = k

            return self.stores.values()

        self.stores = wdict()
        return self.stores.values()

    @_stores.setter
    def _stores(self, what):
        self.stores = what
        if what is not None:
            self.stores = wdict([(s.name, s) for s in what])

    @classmethod
    def get_default(cls, daemon_name):
        db_name = cls.DEFAULT_DB_NAME
        if db_name is None:
            db_name = daemon_name

        return cls(
            debug=True,
            uuid=cls.gen_uuid(),
            secret=cls.gen_secret(),
            name=daemon_name,
            _stores=[
                RelationalStore(
                    name="sql_main",
                    backend="sqlalchemy",
                    pool_size=10,
                    pool_recycle=3600,
                    pool_timeout=30,
                    max_overflow=3,
                    conn_str='postgresql://{user}:@/{daemon}_{user}' \
                                .format(daemon=db_name, user=getpass.getuser()),
                    sync_pool=True,
                    async_pool=True,
                ),
            ],
            main_store='sql_main',
            logger_dest_rotation_period='WEEKLY',
            logger_dest_rotation_compression='gzip',
            _loggers=[
                Logger(
                    path='.',
                    level='DEBUG',
                    # TODO: implement this
                    # format=cls.LOGGING_DEVEL_FORMAT
                ),
            ],
        )

    def apply_storage(self):
        for store in self.stores.values():
            try:
                store.apply()
            except Exception as e:
                logger.exception(e)
                raise

            if self.main_store == store.name:
                engine = store.itself.engine

                import neurons
                neurons.TableModel.Attributes.sqla_metadata.bind = engine

        if self.log_dbconn:
            handler = logging.getLogger().handlers[0]
            _mr = handler._modify_record
            pool = self.get_main_store().engine.pool
            def _modify_record(record):
                _mr(record)
                record.msg = "[%d]%s" % (pool.checkedout(), record.msg)
            handler._modify_record = _modify_record

        return self

    def _parse_overrides(self):
        super(ServiceDaemon, self)._parse_overrides()

        for s in self.stores.values():
            s._parse_overrides()

    def apply(self, daemonize=True):
        """Daemonizes the process if requested, then sets up logging and pid
        files plus data stores.
        """

        # FIXME: apply_storage could return a deferred due to txpool init.

        super(ServiceDaemon, self).apply(daemonize=daemonize)

        self.apply_storage()

        return self

    @staticmethod
    def _set_log_level(ns, level):
        logging.getLogger(ns).setLevel(level)
        logging.info("Sublogger level initialized for '%s' as %s", ns,
                                                        LOGLEVEL_STR_MAP[level])

    RPC_LOGGERS = (
        'spyne.protocol',
        'spyne.protocol.xml',
        'spyne.protocol.dictdoc',
        'spyne.protocol.cloth.to_parent',
        'spyne.protocol.cloth.to_cloth.serializer',

        'neurons.form',
        'neurons.polymer.protocol',
    )

    def pre_logging_apply(self):
        super(ServiceDaemon, self).pre_logging_apply()

        if self.log_protocol or self.log_queries or self.log_results or \
              self.log_model or self.log_interface or self.log_protocol or \
                                                                 self.log_cloth:
            logging.getLogger().setLevel(logging.DEBUG)

        if self.log_model:
            self._set_log_level('spyne.model', logging.DEBUG)
        else:
            self._set_log_level('spyne.model', logging.INFO)

        if self.log_interface:
            self._set_log_level('spyne.interface', logging.DEBUG)
        else:
            self._set_log_level('spyne.interface', logging.INFO)

        if self.log_protocol:
            for ns in self.RPC_LOGGERS:
                self._set_log_level(ns, logging.DEBUG)

        else:
            for ns in self.RPC_LOGGERS:
                self._set_log_level(ns, logging.INFO)

        if self.log_cloth:
            self._set_log_level('spyne.protocol.cloth.to_cloth.cloth',
                                                                  logging.DEBUG)
        else:
            self._set_log_level('spyne.protocol.cloth.to_cloth.cloth',
                                                                   logging.INFO)

        if self.log_sqlalchemy:
            self._set_log_level('sqlalchemy', logging.DEBUG)

        else:
            if self.log_orm:
                self._set_log_level('sqlalchemy.orm', logging.DEBUG)

            if self.log_queries or self.log_results:
                if self.log_queries:
                    self._set_log_level('sqlalchemy.engine', logging.INFO)

                if self.log_results:
                    self._set_log_level('sqlalchemy.engine', logging.DEBUG)

    def get_main_store(self):
        return self.stores[self.main_store].itself

    def get_main_store_config(self):
        return self.stores[self.main_store]



# encoding: utf8
#
# This file is part of the Neurons project.
# Copyright (c), Arskom Ltd. (arskom.com.tr),
#                Burak Arslan <burak.arslan@arskom.com.tr>.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the Arskom Ltd., the neurons project nor the names of
#   its its contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#

from spyne import TTableModel, Integer32
from spyne.model import TTableModelBase
from spyne.store.relational import get_pk_columns

from sqlalchemy.orm import make_transient


class TableModelBase(TTableModelBase()):
    def __init__(self, *args, **kwargs):
        self._changes = set()

        super(TableModelBase, self).__init__(*args, **kwargs)

    def _safe_set(self, key, value, t, attrs):
        retval = super(TableModelBase, self)._safe_set(key, value, t, attrs)
        if retval:
            self._changes.add(key)
        return retval

    @classmethod
    def __respawn__(cls, ctx=None, filters=None):
        has_db = ctx.app is not None and 'sql_main' in ctx.app.config.stores

        if has_db and ctx is not None and ctx.in_object is not None \
                    and len(ctx.in_object) > 0 and ctx.in_object[0] is not None:
            in_object = ctx.in_object[0]

            if filters is None:
                filters = {}

            for k, v in get_pk_columns(cls):
                filters[k] = getattr(in_object, k)

            session = ctx.udc.get_main_session()
            retval = session.query(cls) \
                .with_polymorphic('*') \
                .filter_by(**filters) \
                .all()

            if len(retval) == 0:
                for k, v in get_pk_columns(cls):
                    setattr(in_object, k, None)
                return in_object

            if len(retval) == 1:
                retval = retval[0]

                make_transient(retval)
                for k in in_object._changes:
                    setattr(retval, k, getattr(in_object, k))
                return retval

        if ctx.descriptor.default_on_null:
            return cls.get_deserialization_instance(ctx)


TableModel = TTableModel(base=TableModelBase)
